// WaveFile.cpp: implementation of the WaveFile class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CNY.h"
#include "WaveFile.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

WaveFile::WaveFile()
{
	m_bFileLoaded = false;
}

WaveFile::~WaveFile()
{
	if (m_bFileLoaded)
		FreeData();
}

void WaveFile::FreeData()
{
	free(m_pData);
}

bool WaveFile::ReadFmtChunk(SBFilePtr r_pFile)
{
	// get size of chunk
	DWORD chunkSize;
	r_pFile->ReadData(&chunkSize, sizeof(chunkSize), 1);

	m_format.cbSize = sizeof(m_format);
	// read the WAVEFORMATEX structure
	r_pFile->ReadData(&m_format.wFormatTag, sizeof(m_format.wFormatTag), 1);
	if (m_format.wFormatTag != WAVE_FORMAT_PCM)
	{
		// unsupported format
		r_pFile->Seek(chunkSize-sizeof(m_format.wFormatTag), SEEK_CUR);
		return false;
	}

	// format is ok, so read the rest of the format structure
	r_pFile->ReadData(&m_format.nChannels, sizeof(m_format.nChannels), 1);
	r_pFile->ReadData(&m_format.nSamplesPerSec, sizeof(m_format.nSamplesPerSec), 1);
	r_pFile->ReadData(&m_format.nAvgBytesPerSec, sizeof(m_format.nAvgBytesPerSec), 1);
	r_pFile->ReadData(&m_format.nBlockAlign, sizeof(m_format.nBlockAlign), 1);
	r_pFile->ReadData(&m_format.wBitsPerSample, sizeof(m_format.wBitsPerSample), 1);

	return true;
}

bool WaveFile::ReadDataChunk(SBFilePtr r_pFile)
{
	DWORD chunkSize;
	r_pFile->ReadData(&chunkSize, sizeof(chunkSize), 1);

	// allocate the buffer for WAVE data
	m_pData = malloc(chunkSize);
	m_nDataLength = chunkSize;

	return (r_pFile->ReadData(m_pData, chunkSize, 1) == 1);
}

bool WaveFile::LoadFile(SBFilePtr r_pFile, const String& r_sFileName)
{
	if (m_bFileLoaded)
	{
		FreeData();
		m_bFileLoaded = false;
	}

	if (r_pFile == NULL)
		r_pFile = new DiskFile;

	if (!r_pFile->OpenFile(r_sFileName, FILE_BINARY_R))
	{
		GFC::GetDebugger()->ErrorOut("Unable to open WAV file: " + r_sFileName + "\n");
		return m_bFileLoaded;
	}

	const char riffMagic[4] = { 'R', 'I', 'F', 'F' };
	const char waveMagic[4] = { 'W', 'A', 'V', 'E' };
	const char dataMagic[4] = { 'd', 'a', 't', 'a' };
	const char fmtMagic[4] = { 'f', 'm', 't', ' ' };

	char riffCompare[4];
	char waveCompare[4];
	char chunkCompare[4];

	// check if it is a RIFF file
	if (r_pFile->ReadData(riffCompare, 1, sizeof(riffCompare)) != sizeof(riffCompare)
		|| memcmp(riffCompare, riffMagic, sizeof(riffCompare)) != 0)
	{
		GFC::GetDebugger()->ErrorOut("Not a RIFF file: " + r_sFileName + "\n");
		return m_bFileLoaded;
	}

	// dont care about the file size
	r_pFile->Seek(sizeof(DWORD), SEEK_CUR);

	// check if it is a wave file
	if (r_pFile->ReadData(waveCompare, 1, sizeof(waveCompare)) != sizeof(waveCompare)
		|| memcmp(waveCompare, waveMagic, sizeof(waveCompare)) != 0)
	{
		GFC::GetDebugger()->ErrorOut("Not a WAVE file: " + r_sFileName + "\n");
		return m_bFileLoaded;
	}

	bool fmtLoaded = false;
	bool dataLoaded = false;
	DWORD chunkSize;

	// keep looping for each chunk
	while (true)
	{
		if (r_pFile->ReadData(chunkCompare, 1, sizeof(chunkCompare)) == sizeof(chunkCompare))
		{
			if (memcmp(chunkCompare, fmtMagic, sizeof(chunkCompare)) == 0)
			{
				fmtLoaded = ReadFmtChunk(r_pFile);
			} else if (memcmp(chunkCompare, dataMagic, sizeof(chunkCompare)) == 0)
			{
				dataLoaded = ReadDataChunk(r_pFile);
			} else
			{
				// unknown chunk
				r_pFile->ReadData(&chunkSize, sizeof(DWORD), 1);
				r_pFile->Seek(chunkSize, SEEK_CUR);
			}
		} else
			// reached end of file
			break;
	}

	r_pFile->CloseFile();

	m_bFileLoaded = fmtLoaded && dataLoaded;

	if (!m_bFileLoaded && dataLoaded)
	{
		FreeData();
	}

	if (!m_bFileLoaded)
	{
		GFC::GetDebugger()->ErrorOut("Unsupported WAVE format: " + r_sFileName + "\n");
		return m_bFileLoaded;
	}


	return m_bFileLoaded;
}


void WaveFile::PlayFile()
{
	GetThisApp()->m_aSoundInstances.Insert(new SoundInstance(m_pData, m_nDataLength));
}

