/***************************************
* File    : MultiFile.cpp
* Author  : Shezan Baig 
* Purpose : Implementation of the MultiFile class
***************************************/


#include "stdafx.h"
#include "../Interfaces/SBFile.h"
#include "DiskFile.h"
#include "MultiFile.h"


#include <malloc.h>

IMPLEMENT_RTTI(MultiFile)


// standard constructor
MultiFile::MultiFile()
{
	// add ref to the array of elements
	// to disable to automatic garbage collection
	m_aElements.AddRef();
}

// constructor which will also open the multifile
MultiFile::MultiFile(const String& r_sFileName)
{
	// add ref to the array of elements
	// to disable to automatic garbage collection
	m_aElements.AddRef();

	OpenMultiFile(NULL, r_sFileName);
}

// destructor
MultiFile::~MultiFile()
{
	CloseMultiFile();
}


/**********************************************
* Function        : CreateEmptyMultiFile
* Author          : Shezan Baig 
* Purpose         : creates a file on disk with multifile headers
* Return Value    : true if successful
* Parameters      : String r_sFileName - the multifile to create
**********************************************/
bool MultiFile::CreateEmptyMultiFile(const String& r_sFileName)
{
	DiskFilePtr pFile = new DiskFile;
	if (pFile->OpenFile(r_sFileName, FILE_BINARY_W))
	{
		Gchar magic[4];
		memcpy(magic, "SBMF", sizeof(magic));
		Gint32 fileCnt = 0;
		pFile->WriteData(magic, sizeof(magic), 1);
		pFile->WriteData(&fileCnt, sizeof(fileCnt), 1);
		return true;
	} else
		return false;
}


// , parameters same as fopen()
/**********************************************
* Function        : OpenMultiFile
* Author          : Shezan Baig 
* Purpose         : opens a specified multi-file
* Return Value    : true if successful
* Parameters      : SBFilePtr r_pSBFile - pointer to SBFile interface
*                   String r_sFileName - the multifile to open
**********************************************/
bool MultiFile::OpenMultiFile(SBFilePtr r_pSBFile, const String& r_sFileName)
{
	// close an existing file if it is still opened
	CloseMultiFile();

	m_sFileName = r_sFileName;

	if (r_pSBFile == NULL)
		m_pMultiFile = new DiskFile;
	else
		m_pMultiFile = r_pSBFile;

	if (m_pMultiFile->OpenFile(r_sFileName, FILE_BINARY_R))
	{
		Gchar magic[4];

		if (m_pMultiFile->ReadData(magic, sizeof(magic), 1) != 1)
		{
			m_pMultiFile = NULL;
			return false;
		}

		if (memcmp(magic, "SBMF", sizeof(magic)) == 0)
		{
			int cnt;
			Gint32 nFileCount;
			MultiFileElementPtr felem;

			// get the number of files in the multi-file
			m_pMultiFile->ReadData(&nFileCount, sizeof(nFileCount), 1);
			m_aElements.SetSize(nFileCount);

			// read each of the file elements
			for (cnt = 0; cnt < nFileCount; cnt++)
			{
				felem = new MultiFileElement;
				if (m_pMultiFile->ReadData(&felem->data, sizeof(felem->data), 1) != 1)
				{
					m_pMultiFile = NULL;
					return false;
				}

				m_aElements.Insert(felem);
			}

			return true;
		} else
		{
			m_pMultiFile = NULL;
			return false;
		}
	} else
	{
		m_pMultiFile = NULL;

		GFC::GetDebugger()->DebugOut("Unable to open multi-file: %s\n", (const char*)r_sFileName);
		return false;
	}
}

/**********************************************
* Function        : CloseMultiFile
* Author          : Shezan Baig 
* Purpose         : closes the multi-file
* Return Value    : true if successful
* Parameters      : None
**********************************************/
bool MultiFile::CloseMultiFile()
{
	if (m_pMultiFile != NULL)
	{
		// close an open file (if any) within the multi-file
		CloseFile();

		if (m_pMultiFile->CloseFile())
		{
			// remove all multi-file elements
			m_aElements.SetSize(0);

			m_pMultiFile = NULL;
			return true;
		} else
			return false;
	} else
		return false;
}

/**********************************************
* Function        : IsMultiFileOpen
* Author          : Shezan Baig 
* Purpose         : checks if the multifile is open
* Return Value    : true if it is open
* Parameters      : None
**********************************************/
bool MultiFile::IsMultiFileOpen()
{
	return (m_pMultiFile != NULL);
}


/**********************************************
* Function        : InsertFile
* Author          : Shezan Baig 
* Purpose         : inserts a file into the multifile
*                   pre-conditions: multifile must be opened
*                   post-conditions: any previously opened
*                                    subfiles will be closed,
*                                    multifile will be re-opened
* Return Value    : true if successful
* Parameters      : SBFilePtr r_pFilePtr - pointer to SBFile interface,
*                                          set to NULL to use the
*                                          default class (DiskFile)
*                   String r_sFileName - the file to insert
*                   String r_sInsertAs - the filename to save in the multifile,
*                                        use empty string to use the same name
*                                        as r_sFileName
**********************************************/
bool MultiFile::InsertFile(SBFilePtr r_pFilePtr, const String& r_sFileName, String r_sInsertAs/* = ""*/)
{
	if (m_pMultiFile == NULL)
		return false;

	if (r_sInsertAs.Length() == 0)
		r_sInsertAs = r_sFileName;

	// close any subfile
	CloseFile();

	if (OpenFile(r_sInsertAs, FILE_BINARY_R))
	{
		// the file already exists in the multi-file
		CloseFile();
		throw new Exception("ERROR: Attempted to insert %s into multifile %s, but it already exists.",
							(const char*)r_sInsertAs,
							(const char*)m_sFileName);
		return false;
	}

	// open the file to insert
	SBFilePtr pInsert;
	if (r_pFilePtr == NULL)
		pInsert = new DiskFile;
	else
		pInsert = r_pFilePtr;

	if (!pInsert->OpenFile(r_sFileName, FILE_BINARY_R))
		return false;

	size_t insert_size = pInsert->GetFileSize();
	if (insert_size == 0)
		// don't add zero-byte files
		return false;

	// open a temporary file
	DiskFilePtr pTemp = new DiskFile;
	if (!pTemp->OpenFile(m_sFileName + (String)".tmp", FILE_BINARY_W))
		return false;


	////////////////////////
	// write all the data
	////////////////////////

	// temporary multifile element to store
	// current elements and their offsets
	MultiFileElementPtr felem = new MultiFileElement;

	// write the magic
	Gchar magic[4];
	memcpy(magic, "SBMF", sizeof(magic));
	pTemp->WriteData(magic, sizeof(magic), 1);


	// get the number of files currently in the multi-file
	m_aElements.RemoveItem(NULL);
	Gint32 nFileCount = m_aElements.GetSize();

	// increment nFileCount for the new file
	nFileCount++;
	pTemp->WriteData(&nFileCount, sizeof(nFileCount), 1);

	// default starting position for the new file
	int start_pos = sizeof(magic)
						+ sizeof(nFileCount)
						+ sizeof(felem->data);

	// loop thru current file elements
	int cnt;
	for (cnt = 0; cnt < m_aElements.GetSize(); cnt++)
	{
		*felem = *m_aElements[cnt];
		// increment felem to make room for the new
		// file element
		felem->data.m_nStartPos += sizeof(felem->data);
		felem->data.m_nEndPos += sizeof(felem->data);

		// write the file element
		pTemp->WriteData(&felem->data, sizeof(felem->data), 1);

		start_pos = felem->data.m_nEndPos+1;
	}

	// write the new file element
	felem->data.m_nStartPos = start_pos;
	felem->data.m_nEndPos = felem->data.m_nStartPos
								+ pInsert->GetFileSize() - 1;
	strcpy(felem->data.m_sFileName, r_sInsertAs);
	pTemp->WriteData(&felem->data, sizeof(felem->data), 1);

	// loop thru each of the file elements and
	// write each of the subfiles
	void* filebuffer;
	for (cnt = 0; cnt < m_aElements.GetSize(); cnt++)
	{
		felem = m_aElements[cnt];
		if (!OpenFile(felem->data.m_sFileName, FILE_BINARY_R))
		{
			// failed to open the file
			throw new Exception("ERROR:\n\nFailed to open sub-file in multi-file when attemping to insert new file.\n\nSub-file:%s\nMulti-file:%s",
									felem->data.m_sFileName,
									(const char*)m_sFileName);
			return false;
		}

		filebuffer = malloc(GetFileSize());

		// read from the subfile
		if (ReadData(filebuffer, GetFileSize(), 1) != 1)
		{
			// failed to read data
			free(filebuffer);
			throw new Exception("ERROR:\n\nFailed to read sub-file in multi-file when attemping to insert new file.\n\nSub-file:%s\nMulti-file:%s",
									felem->data.m_sFileName,
									(const char*)m_sFileName);
			return false;
		}

		// write it into the temp file
		pTemp->WriteData(filebuffer, GetFileSize(), 1);

		free(filebuffer);
	}

	// now write pInsert into pTemp
	filebuffer = malloc(insert_size);

	// read from pInsert
	if (pInsert->ReadData(filebuffer, insert_size, 1) != 1)
	{
		// failed to read data
		free(filebuffer);
		throw new Exception("ERROR:\n\nFailed to read input file when attemping to insert file into a multi-file.\n\nInput-file:%s\nMulti-file:%s",
								(const char*)r_sFileName,
								(const char*)m_sFileName);
		return false;
	}

	// write it into the temp file
	pTemp->WriteData(filebuffer, insert_size, 1);

	free(filebuffer);

	pInsert->CloseFile();

	////////////////////////
	// finished write all the data
	////////////////////////

	// close the temporary file and multi-file
	pTemp->CloseFile();
	CloseMultiFile();

	// rename the temporary file to the multi-file
	remove(m_sFileName);
	rename(m_sFileName + (String)".tmp", m_sFileName);

	// re-open the multi-file
	return OpenMultiFile(m_pMultiFile, m_sFileName);
}

/**********************************************
* Function        : ExtractFile
* Author          : Shezan Baig 
* Purpose         : extracts the currently open file
*                   and writes it into r_sFileName
* Return Value    : true if successful
* Parameters      : String r_sFileName - the output file
**********************************************/
bool MultiFile::ExtractToFile(const String& r_sFileName)
{
	if (m_pCurrentFile == NULL)
		return false;

	Seek(0L, SEEK_SET);
	void* buf = malloc(GetFileSize());
	if (ReadData(buf, GetFileSize(), 1) != 1)
	{
		free(buf);
		GFC::GetDebugger()->ErrorOut("Error while reading data from multifile.\n");
		return false;
	}

	DiskFilePtr p = new DiskFile;
	p->OpenFile(r_sFileName, FILE_BINARY_W);
	p->WriteData(buf, GetFileSize(), 1);

	free(buf);

	return true;
}



//////////////////////////////////////////////////////////////////////////////////////////
// Methods inherited from SBFile interface
//////////////////////////////////////////////////////////////////////////////////////////

// opens a file, returns true if successful
// parameters are same as fopen
bool MultiFile::OpenFile(String r_sFileName, String r_sMode)
{
	// close the current file (if any)
	CloseFile();


	if (m_pMultiFile == NULL)
		return false;

	// files cannot be opened for writing
	if (r_sMode == (String)"w"
		|| r_sMode == (String)"a"
		|| r_sMode == (String)"at"
		|| r_sMode == (String)"ab"
		|| r_sMode == (String)FILE_ASCII_W
		|| r_sMode == (String)FILE_ASCII_RW
		|| r_sMode == (String)FILE_BINARY_W
		|| r_sMode == (String)FILE_BINARY_RW)
	{
		throw new Exception("ERROR:\n\nA file element from a multi-file was opened for writing! This is not allowed by the GFC framework.\n\nMulti-File filename: %s",
								(const char*)m_sFileName);
		return false;
	}

	// look for the element
	int cnt;
	MultiFileElementPtr felem;
	for (cnt = m_aElements.GetSize()-1; cnt >= 0; cnt--)
	{
		
		if (stricmp((felem = m_aElements[cnt])->data.m_sFileName, r_sFileName) == 0)
		{
			// found the file, move to the start location
			if (m_pMultiFile->Seek(felem->data.m_nStartPos, SEEK_SET))
			{
				m_pCurrentFile = felem;
				m_nFilePos = 0;
				return true;
			} else
				return false;
		}
	}

	// the file element was not found
	return false;
}

	// closes the file, returns true if successful
bool MultiFile::CloseFile()
{
	if (m_pCurrentFile != NULL)
	{
		m_pCurrentFile = NULL;
		return true;
	} else
		return false;
}

// returns true if a subfile is opened
bool MultiFile::IsFileOpened()
{
	return (m_pCurrentFile != NULL);
}

// reads data from the file
// parameters similar to fread
size_t MultiFile::ReadData(void* buffer, size_t size, size_t count)
{
	if (m_pCurrentFile == NULL || size == 0)
		return 0;

	// calculate possible number of records
	size_t maxRec = (m_pCurrentFile->data.m_nEndPos+1-
					(m_pCurrentFile->data.m_nStartPos+m_nFilePos))/size;

	if (maxRec < count)
		count = maxRec;

	// read from the multi-file
	size_t result = m_pMultiFile->ReadData(buffer, size, count);


	// update file position
	m_nFilePos += result*size;

	// return the number of records read
	return result;
}


// writes data to the file
// parameters similar to fwrite
size_t MultiFile::WriteData(const void* buffer, size_t size, size_t count)
{
	if (m_pCurrentFile == NULL)
		return 0;

	throw new Exception("ERROR:\n\nAttempted to write into a file element from a multi-file! This is not allowed by the GFC framework.\n\nMulti-File filename: %s",
							(const char*)m_sFileName);
	return 0;
}


// returns the size of the file
// (file must be opened first)
size_t MultiFile::GetFileSize()
{
	if (m_pCurrentFile == NULL)
		return 0;
	else
		return m_pCurrentFile->data.m_nEndPos
					-m_pCurrentFile->data.m_nStartPos
					+1;
}

// returns true if file pointer is at the end of the file
bool MultiFile::EndOfFile()
{
	if (m_pCurrentFile == NULL)
		return true;

	return (m_pCurrentFile->data.m_nStartPos
				+m_nFilePos > m_pCurrentFile->data.m_nEndPos);
}

// moves to a position in the file
// parameters similar to fseek()
bool MultiFile::Seek(long offset, int origin)
{
	if (m_pCurrentFile == NULL)
		return false;

	// handle some special cases
	if (origin == SEEK_END)
	{
		if (offset > 0)
		{
			// cannot go forward past the end of the file,
			// so just put the file pointer at the end
			// of the file
			if (m_pMultiFile->Seek(m_pCurrentFile->data.m_nEndPos + 1
										-(m_pCurrentFile->data.m_nStartPos
											+m_nFilePos),
									SEEK_CUR))
			{
				m_nFilePos = m_pCurrentFile->data.m_nEndPos + 1
								-m_pCurrentFile->data.m_nStartPos;
				return true;
			} else
				return false;
		} else if (offset < -((long)GetFileSize()))
		{
			// cannot go backwards past the beginning of the file,
			// so just put the file pointer at the beginning
			// of the file
			if (m_pMultiFile->Seek(-m_nFilePos, SEEK_CUR))
			{
				m_nFilePos = 0;
				return true;
			} else
				return false;
		}
	} else if (origin == SEEK_SET)
	{
		if (offset < 0)
		{
			// cannot go backwards past the beginning of the file,
			// so just put the file pointer at the beginning
			// of the file
			if (m_pMultiFile->Seek(-m_nFilePos, SEEK_CUR))
			{
				m_nFilePos = 0;
				return true;
			} else
				return false;
		} else if (offset > (long)GetFileSize())
		{
			// cannot go forward past the end of the file,
			// so just put the file pointer at the end
			// of the file
			if (m_pMultiFile->Seek(m_pCurrentFile->data.m_nEndPos + 1
										-(m_pCurrentFile->data.m_nStartPos
											+m_nFilePos),
									SEEK_CUR))
			{
				m_nFilePos = m_pCurrentFile->data.m_nEndPos + 1
								-m_pCurrentFile->data.m_nStartPos;
				return true;
			} else
				return false;
		}
	} else if (origin == SEEK_CUR)
	{
		if (offset < -m_nFilePos)
		{
			// cannot go backwards past the beginning of the file,
			// so just put the file pointer at the beginning
			// of the file
			if (m_pMultiFile->Seek(-m_nFilePos, SEEK_CUR))
			{
				m_nFilePos = 0;
				return true;
			} else
				return false;
		} else if (offset > ((long)(GetFileSize()-m_nFilePos)))
		{
			// cannot go forward past the end of the file,
			// so just put the file pointer at the end
			// of the file
			if (m_pMultiFile->Seek(m_pCurrentFile->data.m_nEndPos + 1
										-(m_pCurrentFile->data.m_nStartPos
											+m_nFilePos),
									SEEK_CUR))
			{
				m_nFilePos = m_pCurrentFile->data.m_nEndPos + 1
								-m_pCurrentFile->data.m_nStartPos;
				return true;
			} else
				return false;
		}
	}


	// handled the special cases, now just move the file pointer
	if (origin == SEEK_CUR)
	{
		if (m_pMultiFile->Seek(offset, origin))
		{
			m_nFilePos += offset;
			return true;
		} else
			return false;
	} else if (origin == SEEK_SET)
	{
		if (m_pMultiFile->Seek(m_pCurrentFile->data.m_nStartPos+offset,
								origin))
		{
			m_nFilePos = offset;
			return true;
		} else
			return false;
	} else if (origin == SEEK_END)
	{
		if (m_pMultiFile->Seek(m_pCurrentFile->data.m_nEndPos+offset,
								SEEK_SET))
		{
			m_nFilePos = GetFileSize()+offset;
			return true;
		} else
			return false;
	} else
		return false;
}

// returns the position of the file pointer
// similar to ftell()
long MultiFile::GetPosition()
{
	return m_nFilePos;
}

