/***************************************
* File    : MultiFile.h
* Author  : Shezan Baig 
* Purpose : Declares the MultiFile class
***************************************/

#ifndef _MULTIFILE_H_INCLUDED
#define _MULTIFILE_H_INCLUDED

#include "MultiFileElement.h"

// disable warning related to declspec(dllimport)
// on some Pointer class instatiations
#pragma warning(disable:4251)

/**********************************************
* Class       : MultiFile
* Author      : Shezan Baig 
* Purpose     : Enables multiple virtual files to be
*               read from a single 'pack' file
**********************************************/
class GFCFILES_API MultiFile : public virtual BaseObject,
									public SBFile
{
	DECLARE_RTTI(MultiFile)

	BEGIN_INTERFACE_LIST(MultiFile)
		INTERFACE(SBFile)
		INTERFACE(MultiFile)
	END_INTERFACE_LIST
public:
	// default constructor
	MultiFile();
	// constructor which will also open the multifile
	MultiFile(const String& r_sFileName);
	// destructor
	virtual ~MultiFile();

	// opens a multi-file, parameters same as fopen()
	bool OpenMultiFile(SBFilePtr r_pSBFile, const String& r_sFileName);
	// closes the multi-file
	bool CloseMultiFile();
	// checks if the multifile is open
	bool IsMultiFileOpen();


	// inserts a file into the multifile
	// pre-conditions: multifile must be opened
	// post-conditions: any previously opened subfiles will be closed
	//                  multifile will be re-opened
	bool InsertFile(SBFilePtr r_pFilePtr, const String& r_sFileName, String r_sInsertAs = "");

	// extracts the currently open file
	// and writes it into r_sFileName
	bool ExtractToFile(const String& r_sFileName);


	static bool CreateEmptyMultiFile(const String& r_sFileName);


// Methods inherited from SBFile interface
	// opens a file, returns true if successful
	// parameters are same as fopen
	virtual bool OpenFile(String r_sFileName, String r_sMode);
	// closes the file, returns true if successful
	virtual bool CloseFile();
	// returns true if the file is opened
	virtual bool IsFileOpened();

	// reads data from the file
	// parameters similar to fread
	virtual size_t ReadData(void* buffer, size_t size, size_t count);
	// writes data to the file
	// parameters similar to fwrite
	virtual size_t WriteData(const void* buffer, size_t size, size_t count);


	// returns the size of the file
	// (file must be opened first)
	virtual size_t GetFileSize();
	// returns true if file pointer is at the end of the file
	virtual bool EndOfFile();

	// moves to a position in the file
	// parameters similar to fseek()
	virtual bool Seek(long offset, int origin);
	// returns the position of the file pointer
	// similar to ftell()
	virtual long GetPosition();

private:
	// pointer to the current open file
	MultiFileElementPtr m_pCurrentFile;
	// position of file pointer in m_pCurrentFile
	int m_nFilePos;

	// array of file elements in the multi-file
	MultiFileElementArray m_aElements;

	// file pointer for the entire multi-file
	SBFilePtr m_pMultiFile;
	// filename for multifile
	String m_sFileName;
};
typedef Pointer<MultiFile>		MultiFilePtr;
#pragma warning(default:4251)

#endif // #ifndef _MULTIFILE_H_INCLUDED
