/***************************************
* File    : ConfigReader.cpp
* Author  : Shezan Baig 
* Purpose : Implementation of the ConfigReader class
***************************************/

#include "stdafx.h"

#include "../Global/GFCFramework.h"


IMPLEMENT_RTTI(ConfigReader)

// constructors
ConfigReader::ConfigReader()
: m_pFile(NULL), m_sFileName("")
{
}

ConfigReader::ConfigReader(const String& r_sFileName)
: m_pFile(NULL), m_sFileName(r_sFileName)
{
}


// destructor
ConfigReader::~ConfigReader()
{
	// make sure the file is closed
	CloseFile();
}


/**********************************************
* Function        : ConfigReader::SetFileName
* Author          : Shezan Baig 
* Purpose         : Sets the file name of the config file
* Return Value    : None
* Parameters      : const String& r_sFileName - the new filename
**********************************************/
inline void ConfigReader::SetFileName(const String& r_sFileName)
{
	bool re_open = CloseFile();
	m_sFileName = r_sFileName;
	if (re_open)
		OpenFile();
}

/**********************************************
* Function        : ConfigReader::OpenFile
* Author          : Shezan Baig 
* Purpose         : Opens the configuration file
* Return Value    : bool - true if successful
* Parameters      : None
**********************************************/
inline bool ConfigReader::OpenFile()
{
	if (IsFileOpened())
		return true;

	if (m_sFileName.Length() == 0)
		return false;

	m_pFile = fopen(m_sFileName, "r");

	return IsFileOpened();
}

/**********************************************
* Function        : ConfigReader::CloseFile
* Author          : Shezan Baig 
* Purpose         : Close the configuration file
* Return Value    : bool - false if the file is already closed
* Parameters      : None
**********************************************/
inline bool ConfigReader::CloseFile()
{
	if (IsFileOpened())
	{
		fclose(m_pFile);
		m_pFile = NULL;
		return true;
	} else
		return false;
}

/**********************************************
* Function        : ConfigReader::IsFileOpened
* Author          : Shezan Baig 
* Purpose         : Check if the file is opened
* Return Value    : bool - true if it is opened
* Parameters      : None
**********************************************/
inline bool ConfigReader::IsFileOpened() const
{
	return (m_pFile != NULL);
}


/**********************************************
* Function        : ConfigReader::ReadKey
* Author          : Shezan Baig 
* Purpose         : Read a key's value from the config file
* Return Value    : bool - true if the key was found
* Parameters      : const String& r_sKey - the key to look for
*                   String& r_sValue - the value found
*
* If the key is not found in the configuration file, r_sValue will be
* set to an empty string and the method will return false
**********************************************/
bool ConfigReader::ReadKey(const String& r_sKey, String& r_sValue)
{
	bool bNeedToOpenAndClose = !IsFileOpened();

	if (bNeedToOpenAndClose && !OpenFile())
		// unable to open the file
		return false;

	char lpszLine[256];
	String lpszKeyEquals = r_sKey + "=";

	// cache length of key
	int nKeyLength = lpszKeyEquals.Length();

	r_sValue = "";

	rewind(m_pFile);
	while (fscanf(m_pFile, "%s\n", lpszLine) == 1)
	{
		if (strncmp(lpszLine, lpszKeyEquals, nKeyLength) == 0)
		{
			r_sValue = (char*)(lpszLine + nKeyLength);
			break;
		}
	}


	if (bNeedToOpenAndClose)
		CloseFile();
	return (r_sValue.Length() > 0);
}

