/***************************************
* File    : Float3.h
* Author  : Shezan Baig 
* Purpose : Declaration of Float3 class
***************************************/

#ifndef _FLOAT3_H_INCLUDED
#define _FLOAT3_H_INCLUDED

/**********************************************
* Class       : Float3
* Author      : Shezan Baig 
* Purpose     : encapsulates 3 float values which can be
*               used for vertex and vector information
**********************************************/
class GFCFRAMEWORK_API Float3 : public virtual BaseObject
{
public:
	Float3()
	{
		Set(0.0f, 0.0f, 0.0f);
	}
	Float3(const Float3& r_f3)
	{
		Set(r_f3.GetF0(), r_f3.GetF1(), r_f3.GetF2());
	}
	Float3(float f0, float f1, float f2)
	{
		Set(f0, f1, f2);
	}
	Float3(const float* pf)
	{
		Set(pf);
	}
	virtual ~Float3()
	{
	}

	void Set(float f0, float f1, float f2)
	{
		m_f[0] = f0;
		m_f[1] = f1;
		m_f[2] = f2;
	}
	void Set(const float* pf)
	{
		memcpy(m_f, pf, sizeof(m_f));
	}

	void CopyTo(float* pf) const
	{
		memcpy(pf, m_f, sizeof(m_f));
	}

	float GetF0() const
	{
		return m_f[0];
	}
	float GetF1() const
	{
		return m_f[1];
	}
	float GetF2() const
	{
		return m_f[2];
	}

	inline float& operator[](int nIndex)
	{
		return m_f[nIndex];
	}

	inline float operator[](int nIndex) const
	{
		return m_f[nIndex];
	}

	inline const Float3& operator=(const Float3& r)
	{
		if (&r == this)
			return *this;

		Set(r.m_f);
		return *this;
	}

	friend bool operator==(const Float3& l, const Float3& r);
	friend bool operator!=(const Float3& l, const Float3& r);

	friend Float3 operator+(const Float3& l, const Float3& r);
	friend Float3 operator-(const Float3& l, const Float3& r);
	friend Float3 operator*(const Float3& l, const float& r);
	friend Float3 operator*(const Float3& l, const Float3& r);
	friend Float3 operator/(const Float3& l, const Float3& r);
private:
	float m_f[3];
};
typedef Pointer<Float3>		Float3Ptr;
typedef CTPointer<Float3>	Float3CTPtr;


inline bool operator==(const Float3& l, const Float3& r)
{
	return (l.m_f[0] == r.m_f[0]
			&& l.m_f[1] == r.m_f[1]
			&& l.m_f[2] == r.m_f[2]);
}

inline bool operator!=(const Float3& l, const Float3& r)
{
	return (l.m_f[0] != r.m_f[0]
			|| l.m_f[1] != r.m_f[1]
			|| l.m_f[2] != r.m_f[2]);
}

inline Float3 operator+(const Float3& l, const Float3& r)
{
	Float3 ret;
	ret.m_f[0] = l.m_f[0] + r.m_f[0];
	ret.m_f[1] = l.m_f[1] + r.m_f[1];
	ret.m_f[2] = l.m_f[2] + r.m_f[2];
	return ret;
}

inline Float3 operator-(const Float3& l, const Float3& r)
{
	Float3 ret;
	ret.m_f[0] = l.m_f[0] - r.m_f[0];
	ret.m_f[1] = l.m_f[1] - r.m_f[1];
	ret.m_f[2] = l.m_f[2] - r.m_f[2];
	return ret;
}

inline Float3 operator*(const Float3& l, const float& r)
{
	Float3 ret;
	ret.m_f[0] = l.m_f[0] * r;
	ret.m_f[1] = l.m_f[1] * r;
	ret.m_f[2] = l.m_f[2] * r;
	return ret;
}

inline Float3 operator*(const Float3& l, const Float3& r)
{
	Float3 ret;
	ret.m_f[0] = l.m_f[0] * r.m_f[0];
	ret.m_f[1] = l.m_f[1] * r.m_f[1];
	ret.m_f[2] = l.m_f[2] * r.m_f[2];
	return ret;
}

inline Float3 operator/(const Float3& l, const Float3& r)
{
	Float3 ret;
	if (r.m_f[0] == 0.0f || r.m_f[1] == 0.0f || r.m_f[2] == 0.0f)
	{
		GFC::GetDebugger()->ErrorOut("ERROR: Float3 divide by zero!\n");
		return ret;
	}

	ret.m_f[0] = l.m_f[0] / r.m_f[0];
	ret.m_f[1] = l.m_f[1] / r.m_f[1];
	ret.m_f[2] = l.m_f[2] / r.m_f[2];
	return ret;
}



#endif // _FLOAT3_H_INCLUDED

