/***************************************
* File    : RTTIClass.h
* Author  : Shezan Baig 
* Purpose : Defines classes for runtime type information
***************************************/

#ifndef _RTTICLASS_H_INCLUDED
#define _RTTICLASS_H_INCLUDED

#include "../Inline/String.inl"



typedef void* (*LP_CREATEOBJ_FUNC)(const char* r_sInterface);
typedef bool (*LP_IMPLEMENTSINTERFACE_FUNC)(const char* r_sInterface);
/**********************************************
* Class       : RTTIClass
* Author      : Shezan Baig 
* Purpose     : Holds runtime type information for a class
*
* Note: No garbage collection implemented, because
*       all RTTIClass objects should be statically declared
**********************************************/
class GFCFRAMEWORK_API RTTIClass
{
public:
	RTTIClass(String r_sClassName, LP_CREATEOBJ_FUNC r_pfCreateObj, LP_IMPLEMENTSINTERFACE_FUNC r_pfImplementsInterface);
	~RTTIClass();
	static const RTTIClass* FindClass(String r_sClassName);

	// get the name of the class
	String GetName() const
	{
		return m_sClassName;
	}

	// creates an instance of this class
	void* CreateObject(const char* r_sInterface = NULL) const
	{
		return m_pfCreateObj(r_sInterface);
	}

	bool ImplementsInterface(const char* r_sInterface) const
	{
		return m_pfImplementsInterface(r_sInterface);
	}

private:
	static RTTIClass* m_pRoot;
	// next item in linked list
	RTTIClass* m_pNext;
	// name of the class this RTTIClass represents
	String m_sClassName;
	// function to create the object
	LP_CREATEOBJ_FUNC m_pfCreateObj;
	// function to determine whether the class implements an interface
	LP_IMPLEMENTSINTERFACE_FUNC m_pfImplementsInterface;
};


/**********************************************
* Interface   : RTTI
* Author      : Shezan Baig 
* Purpose     : implement this interface if your class
*				provides RTTI
*
* Note : The BaseObject class implements this interface,
*        therefore all subclasses also implement it.
*        However, each class must override GetClass()
*        inside. (see DECLARE_RTTI and IMPLEMENT_RTTI
*        below)
**********************************************/
class GFCFRAMEWORK_API RTTI
{
public:
	virtual const RTTIClass* GetClass() const = 0;
	virtual void* FindInterface(const char* lpszInterface) const = 0;
};



/*******************************************************
* The following declarations are used to enable runtime
* type information for concrete classes
*******************************************************/
#define DECLARE_RTTI(classname)					\
private:										\
	static RTTIClass rtti_for_##classname;		\
	static void* rtti_Create##classname(const char* r_sInterface)	\
	{											\
		if (r_sInterface != NULL)				\
		{										\
			classname* p = new classname;		\
			void* ret = p->FindInterface(r_sInterface);	\
			if (ret == NULL)					\
				delete p;						\
												\
			return ret;							\
		} else									\
			return new classname;				\
	}											\
												\
	static bool rtti_ImplementsInterface##classname(const char* r_sInterface)	\
	{											\
		classname* p = new classname;			\
		void* ret = p->FindInterface(r_sInterface);	\
		delete p;								\
												\
		return (ret != NULL);					\
	}											\
												\
	/* implementation of RTTI interface */		\
public:											\
	virtual const RTTIClass* GetClass() const	\
	{											\
		return &rtti_for_##classname;			\
	}											\


#define DECLARE_RTTI_SINGLE(classname)			\
private:										\
	static RTTIClass rtti_for_##classname;		\
	static void* rtti_Create##classname(const char* r_sInterface)	\
	{											\
		static classname* pOneAndOnly = NULL;	\
		if (r_sInterface != NULL)				\
		{										\
			void* ret;							\
			if (pOneAndOnly == NULL)			\
			{									\
				pOneAndOnly = new classname;	\
				ret = pOneAndOnly->FindInterface(r_sInterface);	\
				if (ret == NULL)				\
				{								\
					delete pOneAndOnly;			\
					pOneAndOnly = NULL;			\
				}								\
			} else								\
			{									\
				ret = pOneAndOnly->FindInterface(r_sInterface);	\
			}									\
			return ret;							\
		} else									\
		{										\
			pOneAndOnly = new classname;		\
			return pOneAndOnly;					\
		}										\
	}											\
												\
	static bool rtti_ImplementsInterface##classname(const char* r_sInterface)	\
	{											\
		classname* p = new classname;			\
		void* ret = p->FindInterface(r_sInterface);	\
		delete p;								\
												\
		return (ret != NULL);					\
	}											\
												\
	/* implementation of RTTI interface */		\
public:											\
	virtual const RTTIClass* GetClass() const	\
	{											\
		return &rtti_for_##classname;			\
	}											\


#define IMPLEMENT_RTTI(classname)				\
	RTTIClass classname::rtti_for_##classname(""#classname"", classname::rtti_Create##classname, classname::rtti_ImplementsInterface##classname);	\

#define BEGIN_INTERFACE_LIST(classname)											\
		virtual void* FindInterface(const char* lpszInterface) const	\
		{																\
			void* test = NULL;											\
			const char* t_sClassname = ""#classname"";					\

#define INTERFACE(interfacename)		\
			if (strcmp(""#interfacename"", lpszInterface) == 0) \
			{													\
				return (interfacename*)this;					\
			} else if (strcmp(""#interfacename"", t_sClassname) != 0)	\
			{															\
				test = interfacename::FindInterface(lpszInterface);		\
				if (test)										\
					return test;								\
			}													\

#define INTERFACE_EXTERNAL(interfacename)		\
			if (strcmp(""#interfacename"", lpszInterface) == 0) \
			{													\
				return (interfacename*)this;					\
			}													\


#define END_INTERFACE_LIST				\
			return NULL;				\
		}								\




#endif // #ifndef _RTTICLASS_H_INCLUDED
