/***************************************
* File    : Static_GFC.cpp
* Author  : Shezan Baig 
* Purpose : implementation of the GFC class
***************************************/

#include "StdAfx.h"
#include "Static_GFC.h"
#include "ClassNotFoundException.h"

#include "FrameworkState.h"

// global variable about framework state
FRAMEWORK_STATE	state;

GFCApplication* GFC::GetApplication()
{
	return state.m_pApp;
}

DebuggerCTPtr GFC::GetDebugger()
{
	return state.m_pDebugger;
}

void GFC::SetDebugger(DebuggerPtr r_pDebugger)
{
	state.m_pDebugger = r_pDebugger;
}


const String& GFC::GetCommandLine()
{
	return state.m_sCommandLine;
}

void GFC::SetCommandLine(const String& r_sCommandLine)
{
	state.m_sCommandLine = r_sCommandLine;
}

/**********************************************
* Function        : GFC::LoadLibrary
* Author          : Shezan Baig 
* Purpose         : Loads a library into the framework
*                   If the library is already loaded, the
*                   function will just return true
* Return Value    : (bool) - true if successful
* Parameters      : (const String& r_strFileName) - file to load
**********************************************/
bool GFC::LoadLibrary(const String& r_strFileName)
{
	// check if library is already loaded
	LoadedLibraryPtr p;
	for (int i = state.m_arrLoadedLibraries.GetSize()-1; i >= 0; i--)
	{
		p = state.m_arrLoadedLibraries[i];

		if (p != NULL)
		{
			if (p->GetFileName() == r_strFileName)
				// its already loaded, return false
				return true;
		}
	}

	// its not already loaded, so create a new
	// loaded library object
	p = new LoadedLibrary;
	if (p->Load(r_strFileName))
	{
		state.m_arrLoadedLibraries.Insert(p);
		return true;
	} else
		// unable to load
		return false;
}

/**********************************************
* Function        : GFC::UnloadLibrary
* Author          : Shezan Baig 
* Purpose         : Unloads a library from the framework
*                   If the library is not loaded, the function
*                   just exits
* Return Value    : None
* Parameters      : (const String& r_strFileName) - file to unload
**********************************************/
void GFC::UnloadLibrary(const String& r_strFileName)
{
	// look for the loaded library
	LoadedLibraryPtr p;
	for (int i = state.m_arrLoadedLibraries.GetSize()-1; i >= 0; i--)
	{
		p = state.m_arrLoadedLibraries[i];
		if (p != NULL)
		{
			if (p->GetFileName() == r_strFileName)
			{
				// found, delete the LoadedLibrary object
				state.m_arrLoadedLibraries.SetAt(i, NULL);
				return;
			}
		}
	}
}

#include "ConfigReader.h"

/**********************************************
* Function        : GFC::LoadPlatform
* Author          : Shezan Baig 
* Purpose         : Reads a configuration file to load the objects
*                   for the window, input and timer interfaces
* Return Value    : bool - true if all interfaces loaded successfully
* Parameters      : const String& r_sConfigFile - the configuration
*                                                 file used to load
*                                                 the interfaces
**********************************************/
bool GFC::LoadPlatform(const String& r_sConfigFile)
{
	ConfigReader cr(r_sConfigFile);
	cr.AddRef();

	GFC::GetDebugger()->DebugOut("Opening configuration file: %s\n", (const char*)r_sConfigFile);
	if (cr.OpenFile() == false)
	{
		GFC::GetDebugger()->ErrorOut("Unable to load config file: %s\n", (const char*)r_sConfigFile);
		return false;
	}

	char* ps;
	char s[256];
	String sValue;


	// get window interface
	if (cr.ReadKey("Window", sValue) == false)
	{
		GFC::GetDebugger()->ErrorOut("Unable to read Window interface key\n");
		return false;
	}

	strcpy(s, sValue);
	ps = s;

	while (*ps != '\0')
	{
		if (*ps == ',')
		{
			*ps = '\0';
			ps++;

			if (!GFC::LoadLibrary(s))
				return false;

			try
			{
				state.m_pWindow = (Window*)RTTIClass::FindClass(ps)->CreateObject("Window");
			} catch(...)
			{
				GFC::GetDebugger()->ErrorOut("ERROR while creating %s object.\n", ps);
				return false;
			}
			break;
		}
		ps++;
	}

	// get Input interface
	if (cr.ReadKey("Input", sValue) == false)
	{
		GFC::GetDebugger()->ErrorOut("Unable to read Input interface key\n");
		return false;
	}

	strcpy(s, sValue);
	ps = s;

	while (*ps != '\0')
	{
		if (*ps == ',')
		{
			*ps = '\0';
			ps++;

			if (!GFC::LoadLibrary(s))
				return false;

			try
			{
				state.m_pInput = (Input*)RTTIClass::FindClass(ps)->CreateObject("Input");
			} catch(...)
			{
				GFC::GetDebugger()->ErrorOut("ERROR while creating %s object.\n", ps);
				return false;
			}
			break;
		}
		ps++;
	}

	// get Timer interface
	if (cr.ReadKey("Timer", sValue) == false)
	{
		GFC::GetDebugger()->ErrorOut("Unable to read Timer interface key\n");
		return false;
	}

	strcpy(s, sValue);
	ps = s;

	while (*ps != '\0')
	{
		if (*ps == ',')
		{
			*ps = '\0';
			ps++;

			if (!GFC::LoadLibrary(s))
				return false;

			try
			{
				state.m_pTimer = (Timer*)RTTIClass::FindClass(ps)->CreateObject("Timer");
			} catch(...)
			{
				GFC::GetDebugger()->ErrorOut("ERROR while creating %s object.\n", ps);
				return false;
			}
			break;
		}
		ps++;
	}

	GFC::GetDebugger()->DebugOut("Closing configuration file.\n");
	cr.CloseFile();

	return (state.m_pWindow != NULL && state.m_pInput != NULL && state.m_pTimer != NULL);
}

/* the following functions return pointers to the interface timers */
WindowPtr GFC::GetWindow()
{
	return state.m_pWindow;
}

InputPtr GFC::GetInput()
{
	return state.m_pInput;
}

TimerPtr GFC::GetTimer()
{
	return state.m_pTimer;
}

/**********************************************
* Function        : GFC::CleanUp
* Author          : Shezan Baig 
* Purpose         : makes sure the order of cleaning up the framework
*                   is correct, eg all objects removed before
*                   the libraries
* Return Value    : None
* Parameters      : None
**********************************************/
void GFC::CleanUp()
{
	GFC::GetDebugger()->DebugOut("Releasing platform interfaces...\n");
	state.m_pWindow = NULL;
	state.m_pTimer = NULL;
	state.m_pInput = NULL;

	GFC::GetDebugger()->DebugOut("Releasing loaded libraries...\n");
	for (int i = state.m_arrLoadedLibraries.GetSize()-1; i >= 0; i--)
		state.m_arrLoadedLibraries[i] = NULL;

	/* don't release the debugger, let it get released automatically
	   as the stack unwinds
	*/

	GFC::GetDebugger()->DebugOut("CleanUp completed.\n");
}

