/***************************************
* File    : GFCMain.cpp
* Author  : Shezan Baig 
* Purpose : Defines entry point for various
*			types of environments.
*			Also performs GFCApplication
*			initialisation.
***************************************/

#include "StdAfx.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/**********************************************
* Function        : GFCMain
* Author          : Shezan Baig 
* Purpose         : This function is called from main() and WinMain()
*					performs the main application logic, regardless
*					whether its a console app or Win32 app
* Return Value    : always zero
* Parameters      : none
**********************************************/
int GFCMain()
{
	srand(time(NULL));
	if (GFC::LoadPlatform("platform.cfg"))
	{
		GFCApplication* pApp = GFC::GetApplication();
		if (pApp->Init())
		{
			GFC::GetDebugger()->DebugOut("Entering message loop...\n");
			pApp->Run();
			GFC::GetDebugger()->DebugOut("Exited message loop.\n");
		}

		pApp->OnExit();
	} else
	{
		GFC::GetDebugger()->ErrorOut("FAILED to load platform. Unable to continue!\n");
	}

	GFC::CleanUp();

	return 0;
}


/**********************************************
* Class       : DefaultConsoleDebugger
* Author      : Shezan Baig 
* Purpose     : Default implementation of Debugger
*				interface for console applications
**********************************************/
class DefaultConsoleDebugger : public virtual BaseObject, public Debugger
{
	DECLARE_RTTI(DefaultConsoleDebugger)

	BEGIN_INTERFACE_LIST(DefaultColsoleDebugger)
		INTERFACE(Debugger)
		INTERFACE(DefaultConsoleDebugger)
	END_INTERFACE_LIST
public:
	DefaultConsoleDebugger()
	{
	}
	virtual ~DefaultConsoleDebugger()
	{
	}

// Implementation of Debugger
	virtual void DebugOut(const char* lpszString, ...) const
	{
		va_list vl;
		va_start(vl, lpszString);

		vfprintf(stderr, lpszString, vl);

		va_end(vl);
	}
	virtual void ErrorOut(const char* lpszString, ...) const
	{
		va_list vl;
		va_start(vl, lpszString);

		vfprintf(stderr, lpszString, vl);

		va_end(vl);
	}
};

IMPLEMENT_RTTI(DefaultConsoleDebugger)


#include <direct.h>
/**********************************************
* Function        : InitWorkingDirectory
* Author          : Shezan Baig 
* Purpose         : This function sets the working directory
*					to the directory of the application module file
* Return Value    : success or not
* Parameters      : none
**********************************************/
bool InitWorkingDirectory()
{
	char sModule[MAX_PATH];
	char* p;
	GetModuleFileName(NULL, sModule, sizeof(sModule));

	p = sModule;
	while (*p != '\0')
		p++;

	p--;
	while (p > sModule && *p != '\\' && *p != '/')
		p--;

	if (p != sModule)
	{
		*p = '\0';
		return !(_chdir(sModule));
	} else
		return false;
}

/**********************************************
* Function        : main
* Author          : Shezan Baig 
* Purpose         : Entry point for console applications
* Return Value    : (int) Success
* Parameters      : (int) args - number of arguments
*					(char*[]) argv - arguments
**********************************************/
int main(int argc, char* argv[])
{
	try
	{
		InitWorkingDirectory();
		String t_sCommandLine;
		for (int i = 1; i <= argc; i++)
		{
			t_sCommandLine = t_sCommandLine + " " + argv[i];
		}
		GFC::SetCommandLine(t_sCommandLine);
		GFC::SetDebugger(new DefaultConsoleDebugger);
		return GFCMain();
	} catch (Exception* pe)
	{
		printf("\n\nException caught by GFC. Exception type: %s\nProgram will now terminate. Exception description follows.\n\n%s\n\n\nPress ENTER to continue.\n", pe->GetClass()->GetName(), pe->GetDescription());
		delete pe;

		getchar();

		GFC::CleanUp();

		return 1;
	} catch(...)
	{
		GFCApplication* pApp = GFC::GetApplication();
		if (pApp != NULL)
		{
			printf("\n\nUnhandled error in application: %s\nPlease report this error. Vendor information follows.\n\nAuthors:\n%s\n\nCompany Name: %s\n\nContact:\n%s\n\n",
				pApp->GetName(), pApp->GetAuthors(), pApp->GetCompany(), pApp->GetContactInfo());

			// If this is a debug release, rethrow the exception
			// to the C runtime
			#ifdef _DEBUG
			throw;
			#endif // #ifdef _DEBUG
		} else
			printf("Error: GFCApplication object not declared! Please create a global GFCApplication object on the stack. See GFCApplication.h for more information.\n");

		GFC::CleanUp();
		return 1;
	}
}


#ifdef WIN32

/**********************************************
* Class       : DefaultWin32Debugger
* Author      : Shezan Baig 
* Purpose     : Default implementation of Debugger
*				interface for Win32 applications
**********************************************/
class DefaultWin32Debugger : public virtual BaseObject, public Debugger
{
	DECLARE_RTTI(DefaultWin32Debugger)

	BEGIN_INTERFACE_LIST(DefaultWin32Debugger)
		INTERFACE(Debugger)
		INTERFACE(DefaultWin32Debugger)
	END_INTERFACE_LIST
public:
	DefaultWin32Debugger()
	{
	}
	virtual ~DefaultWin32Debugger()
	{
	}

// Implementation of Debugger
	virtual void DebugOut(const char* lpszString, ...) const
	{
		va_list vl;
		va_start(vl, lpszString);

		char s[1024];
		vsprintf(s, lpszString, vl);

		OutputDebugString(s);

		va_end(vl);
	}

	virtual void ErrorOut(const char* lpszString, ...) const
	{
		va_list vl;
		va_start(vl, lpszString);

		char s[1024];
		vsprintf(s, lpszString, vl);

#ifdef _DEBUG
		OutputDebugString(s);
#else // #ifdef _DEBUG
		FILE* f = fopen("errors.txt", "a");
		if (f != NULL)
		{
			fprintf(f, "%s", s);
			fclose(f);
		}
#endif // #ifdef _DEBUG

		va_end(vl);
	}
};
IMPLEMENT_RTTI(DefaultWin32Debugger)

#include <direct.h>
/**********************************************
* Function        : WinMain
* Author          : Shezan Baig 
* Purpose         : entry point for Win32 applications
* Return Value    : (int) Success
* Parameters      : standard WinMain arguments (refer to MSDN)
**********************************************/
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	char s[1024];
	try
	{
		InitWorkingDirectory();

		GFC::SetCommandLine(lpCmdLine);
		GFC::SetDebugger(new DefaultWin32Debugger);

		const char* csCmdLine = GetCommandLine();
		char sCmdLine[1024];
		char* ps;
		if (*csCmdLine == '\"')
		{
			strcpy(sCmdLine, csCmdLine+1);
			ps = sCmdLine;
			while (true)
			{
				if (*ps == '\0')
					break;

				ps++;

				if (*ps == '\"')
				{
					*ps = '\0';
					break;
				}
			}

			// find first slash from back
			while (true)
			{
				if (ps == sCmdLine)
					break;
				if (*ps == '\\' || *ps == '/')
				{
					*ps = '\0';
					break;
				}
				ps--;
			}
			chdir(sCmdLine);
		} else
			strcpy(sCmdLine, csCmdLine);

		return GFCMain();
	} catch (Exception* pe)
	{
		sprintf(s, "Exception caught by GFC. Exception type: %s\nProgram will now terminate. Exception description follows.\n\n%s\n", pe->GetClass()->GetName(), pe->GetDescription());

		MessageBox(HWND_DESKTOP, s, pe->GetClass()->GetName(), MB_ICONSTOP);
		delete pe;

		GFC::CleanUp();
		return 1;
	} catch(...)
	{
		GFCApplication* pApp = GFC::GetApplication();
		if (pApp != NULL)
		{
			sprintf(s, "Unhandled error in application: %s\nPlease report this error. Vendor information follows.\n\nAuthors:\n%s\n\nCompany Name: %s\n\nContact:\n%s\n\n",
				pApp->GetName(), pApp->GetAuthors(), pApp->GetCompany(), pApp->GetContactInfo());
			MessageBox(HWND_DESKTOP, s, "Application Error", MB_ICONSTOP);

			// If this is a debug release, rethrow the exception
			// to the C runtime
			#ifdef _DEBUG
			throw;
			#endif // #ifdef _DEBUG
		} else
			MessageBox(HWND_DESKTOP, "Error: GFCApplication object not declared! Please create a global GFCApplication object on the stack. See GFCApplication.h for more information.\n", "Unable To Initialise GFC", MB_ICONSTOP);

		GFC::CleanUp();
		return 1;
	}
}

#endif // #ifdef WIN32
