/***************************************
* File    : SBArray.inl
* Author  : Shezan Baig 
* Purpose : Array Module.
*			Defines the Array class and
*			also a special pointer which
*			can be used with Array objects
***************************************/


#ifndef _SBARRAY_INL_INCLUDED
#define _SBARRAY_INL_INCLUDED

#include "../Interfaces/Collection.h"

#pragma warning(disable:4275)

/**********************************************
* Class       : class Array<PtrType>
* Author      : Shezan Baig 
* Purpose     : Maintains an array of
*				Pointer<PtrType> objects
*
* Update:
*    Implemented Collection interface in Array class
*       - In() method inserts item at the first available
*         space, or grows the array if necessary
*       - Out() method removes the item at the end
*         of the array
**********************************************/
template <class PtrType>
class INLINE_API Array : public virtual BaseObject,
							public Collection<PtrType>
{
public:
// Construction/Destruction
	Array()
	{
		m_nItemCount = 0;
		m_aItems = NULL;
	}
	Array(int nItemCount)
	{
		m_nItemCount = nItemCount;
		if (nItemCount == 0)
			m_aItems = NULL;
		else
			m_aItems = new PtrType[m_nItemCount];
	}
	Array(const Array<PtrType>& array)
	{
		m_nItemCount = array.m_nItemCount;
		if (m_nItemCount == 0)
		{
			m_aItems = NULL;
		} else
		{
			m_aItems = new PtrType[m_nItemCount];
			for (int i = 0; i < m_nItemCount; i++)
				m_aItems[i] = array.m_aItems[i];
		}
	}

	virtual ~Array()
	{
		if (m_aItems != NULL)
			delete[] m_aItems;
	}

// Operations
	int GetSize() const
	{
		return m_nItemCount;
	}

	void SetSize(int nItemCount)
	{
		if (m_nItemCount == nItemCount)
			return;

		PtrType* temp;
		if (nItemCount == 0)
			temp = NULL;
		else
			temp = new PtrType[nItemCount];

		if (m_nItemCount != 0)
		{
			for (int i = 0; i < nItemCount && i < m_nItemCount; i++)
				temp[i] = m_aItems[i];
			delete[] m_aItems;
		}
		m_nItemCount = nItemCount;
		m_aItems = temp;
	}

	const Array<PtrType>& operator=(const Array<PtrType>& array)
	{
		if (&array == this)
			// don't copy to self
			return *this;

		if (m_aItems != NULL)
			delete[] m_aItems;

		m_nItemCount = array.m_nItemCount;
		if (m_nItemCount == 0)
		{
			m_aItems = NULL;
		} else
		{
			m_aItems = new PtrType[m_nItemCount];
			for (int i = 0; i < m_nItemCount; i++)
				m_aItems[i] = array.m_aItems[i];
		}

		return *this;
	}

	const Array<PtrType>& operator+(const Array<PtrType>& array)
	{
		if (array.GetSize() == 0)
			return *this;

		int i = GetSize();
		int j;
		SetSize(i + array.GetSize());
		for (j = 0; j < array.GetSize(); j++, i++)
			m_aItems[i] = array.m_aItems[j];

		return *this;
	}

	// fills the array with objects using PtrType's default
	// constructor
	void FillDefault()
	{
		for (int i = 0; i < m_nItemCount; i++)
			if (m_aItems[i] == NULL)
				m_aItems[i] = new PtrType;
	}

	// inserts ptr into the array, and grows the array if necessary
	// returns the index where ptr was inserted
	int Insert(PtrType ptr)
	{
		for (int i = 0; i < m_nItemCount; i++)
		{
			if (m_aItems[i] == (const void*)NULL)
			{
				m_aItems[i] = ptr;
				return i;
			}
		}
		SetAtGrow(m_nItemCount, ptr);
		return m_nItemCount-1;
	}

	bool SetAt(int nIndex, PtrType ptr)
	{
		if (nIndex < 0 || nIndex >= m_nItemCount)
			return false;
		m_aItems[nIndex] = ptr;
		return true;
	}

	bool SetAtGrow(int nIndex, PtrType ptr)
	{
		if (nIndex < 0)
			return false;
		if (nIndex >= m_nItemCount)
			SetSize(nIndex+1);
		m_aItems[nIndex] = ptr;
		return true;
	}

	bool InsertAt(int nIndex, PtrType ptr)
	{
		if (nIndex < 0)
			return false;
		SetSize(GetSize()+1);

		// shift all items at nIndex and after nIndex up by one position
		int t_nUpperBound = m_nItemCount-1;
		int i;
		for (i = t_nUpperBound; i > nIndex; i--)
			m_aItems[i] = m_aItems[i-1];

		m_aItems[nIndex] = ptr;
		return true;
	}

	// removes an item at a particular position
	PtrType RemoveAt(int nIndex)
	{
		int i;

		// get the item to return
		PtrType ret = m_aItems[nIndex];

		// shift all items after nIndex down by one position
		int t_nUpperBound = m_nItemCount-1;
		for (i = nIndex; i < t_nUpperBound; i++)
			m_aItems[i] = m_aItems[i+1];

		SetSize(GetSize()-1);

		return ret;
	}

	// removes all items pointing to r_pItem
	// returns the number of items removed
	int RemoveItem(PtrType r_pItem)
	{
		int cnt = 0;
		int cnt1, cnt2;
		for (cnt1 = 0, cnt2 = 0; cnt1 < m_nItemCount; cnt1++)
		{
			if (m_aItems[cnt1] != r_pItem)
			{
				m_aItems[cnt2] = m_aItems[cnt1];
				cnt2++;
			} else
				cnt++;
		}
		SetSize(cnt2);
		return cnt;
	}

	PtrType& GetAt(int nIndex) const
	{
		return m_aItems[nIndex];
	}

	PtrType& operator[](int nIndex) const
	{
		return GetAt(nIndex);
	}

// Methods inherited from Collection interface
	// inserts a pointer to the collection
	virtual void In(const PtrType& p)
	{
		Insert(p);
	}
	// removes a pointer from the collection and returns it
	virtual PtrType Out()
	{
		if (GetSize() == 0)
			return NULL;

		PtrType p = GetAt(GetSize()-1);
		SetSize(GetSize()-1);

		return p;
	}

	// returns a pointer to the next object to be removed by Out()
	virtual PtrType PeekOut() const
	{
		if (GetSize() == 0)
			return NULL;
		else
			return GetAt(GetSize()-1);
	}

	// returns the number of objects in the collection
	virtual int ItemCount() const
	{
		return GetSize();
	}


private:
	int m_nItemCount;
	PtrType* m_aItems;
};


/**********************************************
* Class       : ArrayPointer<PtrType>
* Author      : Shezan Baig 
* Purpose     : A type of Pointer<Array<PtrType>>
*				but contains an additional
*				[] operator for array
*				subscript operation
**********************************************/
template <class PtrType>
class INLINE_API ArrayPointer : public Pointer< Array<PtrType> >
{
public:
	ArrayPointer() : Pointer< Array<PtrType> >()
	{
		ptr = NULL;
	}
	ArrayPointer(const ArrayPointer& r) : Pointer< Array<PtrType> >()
	{
		ptr = r.ptr;
		if (ptr != NULL)
			ptr->AddRef();
	}

	ArrayPointer(Array<PtrType>* init_ptr) : Pointer< Array<PtrType> >()
	{
		ptr = init_ptr;
		if (ptr != NULL)
			ptr->AddRef();
	}

	virtual ~ArrayPointer()
	{
		// ptr will be released in Pointer class
	}

	Pointer<PtrType>& operator[](int nIndex) const
	{
		return ptr->GetAt(nIndex);
	}
};

typedef ArrayPointer<BaseObject>	BaseObjectArray;


#endif // #ifndef _SBARRAY_INL_INCLUDED

