/***************************************
* File    : SBGC.inl
* Author  : Shezan Baig 
* Purpose : Garbage Collection Module.
*			Defines base object class and
*			pointer class
***************************************/

#ifndef _SBGC_INL_INCLUDED
#define _SBGC_INL_INCLUDED

#include "../GFCFramework/RTTIClass.h"


// useless warnings for template instantiation
// (using extern keyword)
#pragma warning(disable:4275)
#pragma warning(disable:4231)

/**********************************************
* Class       : BaseObject
* Author      : Shezan Baig 
* Purpose     : Base class for all classes.
*				All classes/interfaces should 
*				derive from this class in order 
*				to make use of the garbage collection
*
* Note : When deriving from this class, it
*        is recommended to derive publicly
*        and virtually:
*			class A : public virtual BaseObject
*
*        This is to make sure only one copy of
*        BaseObject is created for each instance,
*        especially when inheriting multiple interfaces
*        (Not required if multiple inheritance is not used)
*
*
* IMPORTANT : If objects are created on the stack
*             (i.e. without using the new operator)
*             AddRef() should explicitly be called on that
*             object after it is constructed.
*
*             This is to prevent a premature delete
*             on the object if its address is passed
*             to any Pointer object and released
*             when the pointer dies
**********************************************/
class INLINE_API BaseObject : public RTTI
{
	DECLARE_RTTI(BaseObject)

	BEGIN_INTERFACE_LIST(BaseObject)
	END_INTERFACE_LIST
public:
	BaseObject()
	{
		m_nRefCount = 0;
	}

	virtual ~BaseObject()
	{
	}

	void AddRef()
	{
		m_nRefCount++;
	}
	void Release()
	{
		m_nRefCount--;
		if (m_nRefCount == 0)
			delete this;
	}

	int GetRefCount()
	{
		return m_nRefCount;
	}

private:
	int m_nRefCount;
};




/**********************************************
* Class       : CTPointer<PtrType>
* Author      : Shezan Baig 
* Purpose     : Encapsulates functionality
*				of a constant pointer but
*				also handles the reference
*				count to objects
*
* Note : class PtrType should be directly or
*        indirectly derived from BaseObject
*        This class only stores pointers to single
*        objects, not arrays of objects. Please use
*        the array class for that (see SBArray.inl)
**********************************************/
template <class PtrType>
class INLINE_API CTPointer
{
public:
	// default constructor will set pointer to NULL
	CTPointer()
	{
		ptr = NULL;
	}
	// copy constructor
	CTPointer(const CTPointer<PtrType>& r)
	{
		ptr = r.ptr;
		if (ptr != NULL)
			ptr->AddRef();
	}
	// constructor from standard C pointer,
	// assumes init_ptr is pointing to a single
	// object, not an array of objects
	// for arrays, please use the Array class
	CTPointer(const PtrType* init_ptr)
	{
		ptr = const_cast<PtrType*>(init_ptr);
		if (ptr != NULL)
			ptr->AddRef();
	}

	~CTPointer()
	{
		if (ptr != NULL)
			ptr->Release();
	}

	const CTPointer& operator=(const CTPointer<PtrType>& newptr)
	{
		PtrType* old_ptr = ptr;
		ptr = newptr.ptr;
		if (ptr != NULL)
			ptr->AddRef();
		if (old_ptr != NULL)
			old_ptr->Release();
		return *this;
	}

	// newptr should not point to an array of objects,
	// use Array class for that
	const CTPointer& operator=(const PtrType* newptr)
	{
		PtrType* old_ptr = ptr;
		ptr = const_cast<PtrType*>(newptr);
		if (ptr != NULL)
			ptr->AddRef();
		if (old_ptr != NULL)
			old_ptr->Release();
		return *this;
	}

	operator const PtrType*() const
	{
		return ptr;
	}

	const PtrType* GetObject() const
	{
		return ptr;
	}

	const PtrType* operator->() const
	{
		return ptr;
	}

protected:
	PtrType* ptr;
};

/**********************************************
* Class       : Pointer<PtrType>
* Author      : Shezan Baig 
* Purpose     : Encapsulates functionality
*				of a standard pointer but
*				also handles the reference
*				count to objects
*
* Note : class PtrType should be directly or
*        indirectly derived from BaseObject
*        This class only stores pointers to single
*        objects, not arrays of objects. Please use
*        the array class for that (see SBArray.inl)
**********************************************/
template <class PtrType>
class INLINE_API Pointer
{
public:
	// default constructor will set pointer to NULL
	Pointer()
	{
		ptr = NULL;
	}
	// copy constructor
	Pointer(const Pointer<PtrType>& r)
	{
		ptr = r.ptr;
		if (ptr != NULL)
			ptr->AddRef();
	}
	// constructor from standard C pointer,
	// assumes init_ptr is pointing to a single
	// object, not an array of objects
	// for arrays, please use the Array class
	Pointer(PtrType* init_ptr)
	{
		ptr = init_ptr;
		if (ptr != NULL)
			ptr->AddRef();
	}

	~Pointer()
	{
		if (ptr != NULL)
			ptr->Release();
	}

	const Pointer& operator=(const Pointer<PtrType>& newptr)
	{
		PtrType* old_ptr = ptr;
		ptr = newptr.ptr;
		if (ptr != NULL)
			ptr->AddRef();
		if (old_ptr != NULL)
			old_ptr->Release();
		return *this;
	}

	// newptr should not point to an array of objects,
	// use Array class for that
	const Pointer& operator=(PtrType* newptr)
	{
		PtrType* old_ptr = ptr;
		ptr = newptr;
		if (ptr != NULL)
			ptr->AddRef();
		if (old_ptr != NULL)
			old_ptr->Release();
		return *this;
	}

	operator PtrType*() const
	{
		return ptr;
	}

	PtrType* GetObject() const
	{
		return ptr;
	}

	PtrType* operator->() const
	{
		return ptr;
	}

protected:
	PtrType* ptr;
};



/**********************************************
* Class       : CTPointer<BaseObject>
* Author      : Shezan Baig 
* Purpose     : Instantiation of constant BaseObject
*               Pointer class
*               The private constructor avoids
*               construction of these objects
**********************************************/
class CTPointer<BaseObject>
{
private:
	CTPointer<BaseObject>()
	{
	}
	~CTPointer<BaseObject>()
	{
	}
};

/**********************************************
* Class       : Pointer<BaseObject>
* Author      : Shezan Baig 
* Purpose     : Instantiation of BaseObject
*               Pointer class
*               The private constructor avoids
*               construction of these objects
**********************************************/
class Pointer<BaseObject>
{
private:
	Pointer<BaseObject>()
	{
	}
	~Pointer<BaseObject>()
	{
	}
};



#define STD_INSTANTIATION(api_type, class_name)		\
extern template class api_type Pointer<class_name>;		\
extern template class api_type Array<Pointer<class_name> >;		\



#endif // #ifndef _SBGC_INL_INCLUDED
