/***************************************
* File    : PlatformInterfaces.h
* Author  : Shezan Baig 
* Purpose : Interface definitions for platform independence
***************************************/

#ifndef _PLATFORMINTERFACES_H_INCLUDED
#define _PLATFORMINTERFACES_H_INCLUDED

// constants for nEvent parameter for keyboard function
#define KB_KEYDOWN		1
#define KB_KEYUP		2


// constants for nEvent parameter of mouse function
#define MOUSE_LBUTTONDOWN		1
#define MOUSE_LBUTTONUP			2
#define MOUSE_RBUTTONDOWN		3
#define MOUSE_RBUTTONUP			4
#define MOUSE_MBUTTONDOWN		5
#define MOUSE_MBUTTONUP			6
#define MOUSE_LBUTTONDBLCLK		7
#define MOUSE_RBUTTONDBLCLK		8
#define MOUSE_MBUTTONDBLCLK		9
#define MOUSE_MOVE				10

/**********************************************
* Interface   : InputTarget
* Author      : Shezan Baig 
* Purpose     : Interface for passing input messages
**********************************************/
class InputTarget : public virtual BaseObject
{
public:
	// callback function for keyboard events
	virtual void KeyboardFunction(int nEvent, int nKeyCode) = 0;
	// callback function for mouse events
	virtual void MouseFunction(int nEvent, int nPosX, int nPosY) = 0;
};
typedef Pointer<InputTarget>		InputTargetPtr;

/**********************************************
* Interface     : OGL
* Author        : Shezan Baig 
* Purpose       : interface for OpenGL related functions
**********************************************/
class OGL : public virtual BaseObject
{
public:
	// callback function after creating rendering context
	virtual void OnInitialiseRC() = 0;
	// callback function before deleting rendering context
	virtual void OnDeleteRC() = 0;
	// callback function for drawing scene
	virtual void OnDrawScene() = 0;
	// callback function when window size changes
	virtual void OnWindowSizeChange(int nWidth, int nHeight) = 0;
};
typedef Pointer<OGL>		OGLPtr;


// remove CreateWindow definition from winuser.h
#ifdef CreateWindow
#undef CreateWindow
#endif // #ifdef CreateWindow

/**********************************************
* Interface     : Window
* Author        : Shezan Baig 
* Purpose       : interface for window handling functions
**********************************************/
class Window : public virtual BaseObject
{
public:
	// creates window and initialises OpenGL
	virtual bool CreateWindow(const String& lpszTitle) = 0;
	// close the window and unload OpenGL
	virtual bool CloseWindow() = 0;
	// display the window
	virtual bool ShowWindow() = 0;
	// invalidates the window and forces repaint
	virtual void InvalidateWindow() = 0;

	// set the window to full screen mode
	virtual bool SetFullScreen() = 0;
	// set the window not to full screen
	virtual bool SetWindowRect(int left, int top, int right, int bottom) = 0;

	// process all messages in message loop
	// returns false when application needs to quit
	virtual bool FlushMessages() = 0;

	// sets the pointer to the IOpenGL interface for callback functions
	virtual void SetOpenGL(OGLPtr lpOpenGL) = 0;

	// this function will only work under Win32
	virtual HWND GetHwnd() = 0;
	virtual HDC GetHdc() = 0;

	// this function will only work if OpenGL is setup
	virtual HGLRC GetHglRC() = 0;
	virtual void CaptureRC() = 0;
	virtual void ReleaseRC() = 0;
};
typedef Pointer<Window>				WindowPtr;

/**********************************************
* Interface     : Timer
* Author        : Shezan Baig 
* Purpose       : Provide timing functions
**********************************************/
class Timer : public virtual BaseObject
{
public:
	virtual unsigned long GetTickCount() const = 0;
};
typedef Pointer<Timer>				TimerPtr;


/**********************************************
* Interface     : Input
* Author        : Shezan Baig 
* Purpose       : interface for receiving input
**********************************************/
class Input : public virtual BaseObject
{
public:
	// returns true if nKeyCode is currently down
	virtual bool IsKeyDown(int nKeyCode) const = 0;
	// returns true if left mouse button is currently down
	virtual bool IsLButtonDown() const = 0;
	// returns true if right mouse button is currently down
	virtual bool IsRButtonDown() const = 0;

	// sets the pointer to the IApplication interface for callback functions
	virtual void SetInputTarget(InputTargetPtr lpTarget) = 0;

	// sets the position of the mouse to the ones specified by nX and xY
	// if nX and nY are -1, then the mouse is centered in the window, and
	// nX and nY are updated accordingly
	virtual void SetMousePosition(int& nX, int& nY) = 0;
};
typedef Pointer<Input>			InputPtr;





#endif // _PLATFORMINTERFACES_H_INCLUDED
