/***************************************
* File    : LoadCOB.cpp
* Author  : Shezan Baig 
* Date    : 5/3/2001
* Purpose : implementation for loading COB files
***************************************/

#include "stdafx.h"

#include "SBObject_Types.h"
#include "../Interfaces/SB3DObject.h"
#include "SBObjectClass.h"



void TransformVector(SBvector v, float trans[4][4])
{
	SBvector result;

	result[0] = trans[0][3];
	result[1] = trans[1][3];
	result[2] = trans[2][3];
	for (int i = 0; i < 3; i++)
	{
		result[0] += v[i]*trans[0][i];
		result[1] += v[i]*trans[1][i];
		result[2] += v[i]*trans[2][i];
	}
	v[0] = result[0];
	v[1] = result[1];
	v[2] = result[2];
}


int SBObject::ProcessCOBMat1Chunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug)
{
//	short nStringLen;
	int nStartPos = fptr->GetPosition(); // starting position of this chunk




	int nMaterialCount;
	short nMaterialNo;
	// read the material index number
	if (fptr->ReadData(&nMaterialNo, sizeof(nMaterialNo), 1) != 1)
		return COBCHUNK_ERROR;

	nMaterialCount = nMaterialNo+1;


	if (nMaterialCount < pParentObject->GetMaterialCount())
		nMaterialCount = pParentObject->GetMaterialCount();

	////////////////////////////////////////////////////////////
	// copy existing materials from parent object to
	// temporary buffer
	Tag_Material* t_pMaterials = new Tag_Material[nMaterialCount];
	int counter;
	for (counter = 0; counter < pParentObject->GetMaterialCount(); counter++)
	{
		t_pMaterials[counter] = pParentObject->m_pMaterials[counter];
	}


	///////////////////////////////////////////////
	// load the material from the file


	// advance 3 bytes (for shader type, facet type & autofacet angle)
	char dummy;
	if (fptr->ReadData(&dummy, sizeof(dummy), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&dummy, sizeof(dummy), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&dummy, sizeof(dummy), 1) != 1)
		return COBCHUNK_ERROR;
	

	// read the red, green, blue components
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_rgbaColor[0], sizeof(t_pMaterials[nMaterialNo].m_rgbaColor[0]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_rgbaColor[1], sizeof(t_pMaterials[nMaterialNo].m_rgbaColor[1]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_rgbaColor[2], sizeof(t_pMaterials[nMaterialNo].m_rgbaColor[2]), 1) != 1)
		return COBCHUNK_ERROR;

	// read the opacity
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_rgbaColor[3], sizeof(t_pMaterials[nMaterialNo].m_rgbaColor[3]), 1) != 1)
		return COBCHUNK_ERROR;

	// read ambience
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_fAmbience, sizeof(t_pMaterials[nMaterialNo].m_fAmbience), 1) != 1)
		return COBCHUNK_ERROR;
	t_pMaterials[nMaterialNo].m_fAmbience = (t_pMaterials[nMaterialNo].m_fAmbience*2.0f)-1.0f;
	// read specular
	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_fSpecular, sizeof(t_pMaterials[nMaterialNo].m_fSpecular), 1) != 1)
		return COBCHUNK_ERROR;
	// read roughness
//	if (fptr->ReadData(&t_pMaterials[nMaterialNo].m_fRoughness, sizeof(t_pMaterials[nMaterialNo].m_fRoughness), 1) != 1)
//		return COBCHUNK_ERROR;
//	t_pMaterials[nMaterialNo].m_fRoughness = (t_pMaterials[nMaterialNo].m_fRoughness*0.7f)+0.3f;


	// the rest of the bytes can be ignored
	// (until textures is implemented)



	//////////////////////////////////////////////////////////
	// update the parent object's material buffer
	// with the new data
	pParentObject->SetMaterials(nMaterialCount, t_pMaterials);



	int nEndPos = fptr->GetPosition(); // ending position of this chunk
	if (nEndPos-nStartPos != hdr.size)
	{
		if (pDebug != NULL && nEndPos-nStartPos > hdr.size)
			// we have overshot the chunk, display warning
			pDebug->ErrorOut("ProcessCOBMat1Chunk no: %d - for %s overshot chunk size! End Pos = %d\n", nMaterialNo, pParentObject->m_sName, nEndPos);
		// make sure the next file position is after the entire chunk
		// in case new variables are added in future
		fptr->Seek(hdr.size-(nEndPos-nStartPos), SEEK_CUR);
	}


	return COBCHUNK_CONTINUE;
}

#define F_HOLE		0x08	// (bit 3)

int SBObject::ProcessCOBPolHChunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug)
{
	short nStringLen;
	int nStartPos = fptr->GetPosition(); // starting position of this chunk

	Pointer<SBObject> pObject;
	if (hdr.parent != 0)
	{
		if (pParentObject == NULL)
		{
			if (pDebug != NULL)
				pDebug->ErrorOut("Error in COB file. Unable to find parent %d for object ID %d\n", hdr.parent, hdr.id);
			return COBCHUNK_ERROR;
		}

		// create new object and add it to the parent object
		pObject = new SBObject;
		pObject->m_pParentObject = pParentObject;
		pParentObject->m_aSubObjects.Insert(pObject);
	} else // hdr.parent == 0, assign pObject to the main object
	{
		pObject = pParentObject;
		pObject->m_pParentObject = NULL;
	}


	// add object to object list
	Tag_ObjectIDMap objIDMap;
	objIDMap.m_nID = hdr.id;
	objIDMap.m_pObject = pObject;
	cobContext.m_aObjects.push_back(objIDMap);


	///////// read name //////////
	short nDupeCount;
	char sDupeCount[10];
	// reed dupecount
	if (fptr->ReadData(&nDupeCount, sizeof(nDupeCount), 1) != 1)
		return COBCHUNK_ERROR;
	if (nDupeCount > 0)
	{
		strcpy(sDupeCount, ",");
		itoa(nDupeCount, sDupeCount+1, 10);
	} else
		strcpy(sDupeCount, "");

	// get string length
	if (fptr->ReadData(&nStringLen, sizeof(nStringLen), 1) != 1)
		return COBCHUNK_ERROR;
	// read name
	if (nStringLen != 0)
	{
		pObject->m_sName = new char[nStringLen+strlen(sDupeCount)+1];
		if (fptr->ReadData(pObject->m_sName, 1, nStringLen) != nStringLen)
			return COBCHUNK_ERROR;
		pObject->m_sName[nStringLen] = '\0';
	} else
	{
		nStringLen = 6;
		pObject->m_sName = new char[nStringLen+strlen(sDupeCount)+1];
		strcpy(pObject->m_sName, "NoName");
	}
	strcat(pObject->m_sName, sDupeCount);

	
	////////// read local axes //////////

	// get center of axis
	if (fptr->ReadData(&pObject->m_vPosition, sizeof(pObject->m_vPosition), 1) != 1)
		return COBCHUNK_ERROR;
	// read x,y,z axes direction
	if (fptr->ReadData(&pObject->m_avAxes[SB_X], sizeof(pObject->m_avAxes[SB_X]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&pObject->m_avAxes[SB_Y], sizeof(pObject->m_avAxes[SB_Y]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&pObject->m_avAxes[SB_Z], sizeof(pObject->m_avAxes[SB_Z]), 1) != 1)
		return COBCHUNK_ERROR;


	////////// read current position ///////
	float currentposition[4][4];
	if (fptr->ReadData(&currentposition[0], sizeof(currentposition[0]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&currentposition[1], sizeof(currentposition[1]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&currentposition[2], sizeof(currentposition[2]), 1) != 1)
		return COBCHUNK_ERROR;

	// default coordinates for 4th row of current position
	currentposition[3][0] = 0.0f;
	currentposition[3][1] = 0.0f;
	currentposition[3][2] = 0.0f;
	currentposition[3][3] = 1.0f;


	// offset from parent objects
	Pointer<SBObject> pParentTest = pObject->m_pParentObject;
	while (pParentTest != NULL)
	{
		TranslateVertex(pObject->m_vPosition, -pParentTest->m_vPosition[SB_X], -pParentTest->m_vPosition[SB_Y], -pParentTest->m_vPosition[SB_Z]);
		pParentTest = pParentTest->m_pParentObject;
	}


	/////////  read local vertex list /////////

	// get number of vertices
	int nCount;
	int nCounter;

	if (fptr->ReadData(&nCount, sizeof(nCount), 1) != 1)
		return COBCHUNK_ERROR;
	pObject->SetVerticeCount(nCount);
	for (nCounter = 0; nCounter < nCount; nCounter++)
	{
		if (fptr->ReadData(&pObject->m_pVertices[nCounter], sizeof(SBvertex), 1) != 1)
			return COBCHUNK_ERROR;

		// apply object transformation
		TransformVector(pObject->m_pVertices[nCounter], currentposition);

		TranslateVertex(pObject->m_pVertices[nCounter], -pObject->m_vPosition[SB_X], -pObject->m_vPosition[SB_Y], -pObject->m_vPosition[SB_Z]);

		// offset from parent objects
		pParentTest = pObject->m_pParentObject;
		while (pParentTest != NULL)
		{
			TranslateVertex(pObject->m_pVertices[nCounter], -pParentTest->m_vPosition[SB_X], -pParentTest->m_vPosition[SB_Y], -pParentTest->m_vPosition[SB_Z]);
			pParentTest = pParentTest->m_pParentObject;
		}
	}


	///////// read UV list ////////////

	// get number of coords
	if (fptr->ReadData(&nCount, sizeof(nCount), 1) != 1)
		return COBCHUNK_ERROR;
	pObject->SetTextureCoordCount(nCount);
	for (nCounter = 0; nCounter < nCount; nCounter++)
	{
		if (fptr->ReadData(&pObject->m_pTextures[nCounter], sizeof(pObject->m_pTextures[nCounter]), 1) != 1)
			return COBCHUNK_ERROR;
	}




	/////////// read faces /////////////

	// get number of faces
	if (fptr->ReadData(&nCount, sizeof(nCount), 1) != 1)
		return COBCHUNK_ERROR;

	// variables for reading faces
	unsigned char faceFlag;
	short numVert;

	int numFaces = 0;
	int faceOffset = 0;
	std::vector < Tag_TmpVector, std::allocator<Tag_TmpVector> > t_aNormals;

	// nCount now contains the total number of polygon faces
	// loop through once to get the total number of triangular faces
	for (nCounter = 0; nCounter < nCount; nCounter++)
	{
		if (fptr->ReadData(&faceFlag, sizeof(faceFlag), 1) != 1)
			return COBCHUNK_ERROR;
		if (fptr->ReadData(&numVert, sizeof(numVert), 1) != 1)
			return COBCHUNK_ERROR;
		if (!(faceFlag & F_HOLE))
		{
			// skip the material index
			fptr->Seek(2L, SEEK_CUR);
			faceOffset += 2;
		}
		// skip vertex list and UV list
		fptr->Seek(8L*numVert, SEEK_CUR);

		faceOffset += sizeof(faceFlag) + sizeof(numVert) + 8L*numVert;

		numFaces += numVert - 2;
	}

	// go back to beginning of faces
	fptr->Seek(-faceOffset, SEEK_CUR);

	int nCurFace = 0;
	int nCounter2;
	int nCounterNormal;
	short nMat;
	SBvector vNormal;
	Tag_TmpVector temp;

	pObject->SetFaceCount(numFaces);
	for (nCounter = 0; nCounter < nCount; nCounter++)
	{
		if (fptr->ReadData(&faceFlag, sizeof(faceFlag), 1) != 1)
			return COBCHUNK_ERROR;
		if (fptr->ReadData(&numVert, sizeof(numVert), 1) != 1)
			return COBCHUNK_ERROR;

		if (!(faceFlag & F_HOLE))
			// read the material index
			if (fptr->ReadData(&nMat, sizeof(nMat), 1) != 1)
				return COBCHUNK_ERROR;


		for (nCounter2 = 0; nCounter2 < numVert-2; nCounter2++, nCurFace++)
		{
			pObject->m_pFaces[nCurFace].m_nMaterial = nMat;
			if (nCounter2 == 0)
			{
				// read first 3
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nVerts[0], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nTextureCoords[0], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nVerts[2], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nTextureCoords[2], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nVerts[1], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nTextureCoords[1], 4, 1) != 1)
					return COBCHUNK_ERROR;

				CalcNormal(pObject->m_pVertices[pObject->m_pFaces[nCurFace].m_nVerts[0]],
							pObject->m_pVertices[pObject->m_pFaces[nCurFace].m_nVerts[1]],
							pObject->m_pVertices[pObject->m_pFaces[nCurFace].m_nVerts[2]],
							vNormal);

				for (nCounterNormal = 0; nCounterNormal < t_aNormals.size(); nCounterNormal++)
					if (VectorsEqual(vNormal, t_aNormals[nCounterNormal].v))
						break;

				if (nCounterNormal == t_aNormals.size())
				{
					CopyVector(temp.v, vNormal);
					t_aNormals.push_back(temp); // vNormal wasnt found in t_aNormals
				}
			} else
			{
				// read next
				pObject->m_pFaces[nCurFace].m_nVerts[0] = pObject->m_pFaces[nCurFace-1].m_nVerts[0];
				pObject->m_pFaces[nCurFace].m_nTextureCoords[0] = pObject->m_pFaces[nCurFace-1].m_nTextureCoords[0];

				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nVerts[1], 4, 1) != 1)
					return COBCHUNK_ERROR;
				if (fptr->ReadData(&pObject->m_pFaces[nCurFace].m_nTextureCoords[1], 4, 1) != 1)
					return COBCHUNK_ERROR;

				pObject->m_pFaces[nCurFace].m_nVerts[2] = pObject->m_pFaces[nCurFace-1].m_nVerts[1];
				pObject->m_pFaces[nCurFace].m_nTextureCoords[2] = pObject->m_pFaces[nCurFace-1].m_nTextureCoords[1];
			}
			
			// Default behaviour is to assign the same normal for all
			// vertices for each face, this will result in faceted surfaces
			// call I3DObject::SmoothSurfaces to assign smooth normals
			pObject->m_pFaces[nCurFace].m_nNormals[0] = nCounterNormal;
			pObject->m_pFaces[nCurFace].m_nNormals[1] = nCounterNormal;
			pObject->m_pFaces[nCurFace].m_nNormals[2] = nCounterNormal;
		}
	}

	// allocate normals
	nCount = t_aNormals.size();
	pObject->SetNormalCount(nCount);
	for (nCounter = 0; nCounter < nCount; nCounter++)
		CopyVector(pObject->m_pNormals[nCounter], t_aNormals[nCounter].v);

	// delete all items from t_aNormals
	while (t_aNormals.size() > 0)
		t_aNormals.erase(t_aNormals.begin());


	int nEndPos = fptr->GetPosition(); // ending position of this chunk
	if (nEndPos-nStartPos != hdr.size)
	{
		if (pDebug != NULL && nEndPos-nStartPos > hdr.size)
			// we have overshot the chunk, display warning
			pDebug->ErrorOut("ProcessCOBPolHChunk for %s overshot chunk size! End Pos = %d\n", pObject->m_sName, nEndPos);
		// make sure the next file position is after the entire chunk
		// in case new variables are added in future
		fptr->Seek(hdr.size-(nEndPos-nStartPos), SEEK_CUR);
	}


	return COBCHUNK_CONTINUE;
}


int SBObject::ProcessCOBGrouChunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug)
{
	short nStringLen;
	int nStartPos = fptr->GetPosition(); // starting position of this chunk

	Pointer<SBObject> pObject;
	if (hdr.parent != 0)
	{
		if (pParentObject == NULL)
		{
			if (pDebug != NULL)
				pDebug->ErrorOut("Error in COB file. Unable to find parent %d for object ID %d\n", hdr.parent, hdr.id);
			return COBCHUNK_ERROR;
		}

		// create new object and add it to the parent object
		pObject = new SBObject;
		pObject->m_pParentObject = pParentObject;
		pParentObject->m_aSubObjects.Insert(pObject);
	} else // hdr.parent == 0, assign pObject to the main object
	{
		pObject = pParentObject;
		pObject->m_pParentObject = NULL;
	}


	// add object to object list
	Tag_ObjectIDMap objIDMap;
	objIDMap.m_nID = hdr.id;
	objIDMap.m_pObject = pObject;
	cobContext.m_aObjects.push_back(objIDMap);


	///////// read name //////////
	
	short nDupeCount;
	char sDupeCount[10];
	// reed dupecount
	if (fptr->ReadData(&nDupeCount, sizeof(nDupeCount), 1) != 1)
		return COBCHUNK_ERROR;
	if (nDupeCount > 0)
	{
		strcpy(sDupeCount, ",");
		itoa(nDupeCount, sDupeCount+1, 10);
	} else
		strcpy(sDupeCount, "");

	// get string length
	if (fptr->ReadData(&nStringLen, sizeof(nStringLen), 1) != 1)
		return COBCHUNK_ERROR;
	// read name
	if (nStringLen != 0)
	{
		pObject->m_sName = new char[nStringLen+strlen(sDupeCount)+1];
		if (fptr->ReadData(pObject->m_sName, 1, nStringLen) != nStringLen)
			return COBCHUNK_ERROR;
		pObject->m_sName[nStringLen] = '\0';
	} else
	{
		nStringLen = 6;
		pObject->m_sName = new char[nStringLen+strlen(sDupeCount)+1];
		strcpy(pObject->m_sName, "NoName");
	}
	strcat(pObject->m_sName, sDupeCount);

	
	////////// read local axes //////////

	// get center of axis
	if (fptr->ReadData(&pObject->m_vPosition, sizeof(pObject->m_vPosition), 1) != 1)
		return COBCHUNK_ERROR;
	// read x,y,z axes direction
	if (fptr->ReadData(&pObject->m_avAxes[SB_X], sizeof(pObject->m_avAxes[SB_X]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&pObject->m_avAxes[SB_Y], sizeof(pObject->m_avAxes[SB_Y]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&pObject->m_avAxes[SB_Z], sizeof(pObject->m_avAxes[SB_Z]), 1) != 1)
		return COBCHUNK_ERROR;


	////////// read current position ///////
	float currentposition[4][4];
	if (fptr->ReadData(&currentposition[0], sizeof(currentposition[0]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&currentposition[1], sizeof(currentposition[1]), 1) != 1)
		return COBCHUNK_ERROR;
	if (fptr->ReadData(&currentposition[2], sizeof(currentposition[2]), 1) != 1)
		return COBCHUNK_ERROR;

	// default coordinates for 4th row of current position
	currentposition[3][0] = 0.0f;
	currentposition[3][1] = 0.0f;
	currentposition[3][2] = 0.0f;
	currentposition[3][3] = 1.0f;

	// offset from parent objects
	Pointer<SBObject> pParentTest = pObject->m_pParentObject;
	while (pParentTest != NULL)
	{
		TranslateVertex(pObject->m_vPosition, -pParentTest->m_vPosition[SB_X], -pParentTest->m_vPosition[SB_Y], -pParentTest->m_vPosition[SB_Z]);
		pParentTest = pParentTest->m_pParentObject;
	}


	int nEndPos = fptr->GetPosition(); // ending position of this chunk
	if (nEndPos-nStartPos != hdr.size)
	{
		if (pDebug != NULL && nEndPos-nStartPos > hdr.size)
			// we have overshot the chunk, display warning
			pDebug->ErrorOut("ProcessCOBGrouChunk overshot chunk size! End Pos = %d\n", nEndPos);
		// make sure the next file position is after the entire chunk
		// in case new variables are added in future
		fptr->Seek(hdr.size-(nEndPos-nStartPos), SEEK_CUR);
	}


	return COBCHUNK_CONTINUE;
}


int SBObject::ProcessCOBChunk(SBFilePtr fptr, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, Pointer<SBObject> pObject, DebuggerCTPtr pDebug)
{
	// pObject points to the root object, any chunk where parentid = 0 should refer to this object
	Pointer<SBObject> pParentObject = NULL;
	if (hdr.parent != 0)
	{
		int nCounter;
		for (nCounter = 0; nCounter < cobContext.m_aObjects.size(); nCounter++)
		{
			if (cobContext.m_aObjects[nCounter].m_nID == hdr.parent)
			{
				pParentObject = cobContext.m_aObjects[nCounter].m_pObject;
				break;
			}
		}
	} else
		pParentObject = pObject;


	if (memcmp(hdr.type, "END ", sizeof(hdr.type)) == 0)
	{
		return COBCHUNK_EOF;
	} else if (memcmp(hdr.type, "Grou", sizeof(hdr.type)) == 0)
	{
		return ProcessCOBGrouChunk(fptr, pParentObject, hdr, cobContext, pDebug);
	} else if (memcmp(hdr.type, "PolH", sizeof(hdr.type)) == 0)
	{
		return ProcessCOBPolHChunk(fptr, pParentObject, hdr, cobContext, pDebug);
	} else if (memcmp(hdr.type, "Mat1", sizeof(hdr.type)) == 0)
	{
		return ProcessCOBMat1Chunk(fptr, pParentObject, hdr, cobContext, pDebug);
	} else
	{
		// unknown chunk, skip section
		fptr->Seek(hdr.size, SEEK_CUR);
		return COBCHUNK_UNKNOWN;
	}
}

bool SBObject::LoadCOBFile(SBFilePtr fptr, const String& lpszFileName, DebuggerCTPtr pDebug)
{
	if (fptr->OpenFile(lpszFileName, FILE_BINARY_R))
	{
		Tag_COBHeader hdr;
		Pointer<SBObject> pObject = NULL;

		if (fptr->ReadData(&hdr, sizeof(hdr), 1) == 1)
		{
			if (memcmp(hdr.magic, KNOWN_COB_MAGIC, sizeof(hdr.magic)) == 0)
			{
				if (hdr.format == 'B')
				{
					Tag_COBContext cobContext;
					int read_result;
					int result;
					Tag_COBChunkHeader chunk;
					pObject = this;

					while ((read_result = fptr->ReadData(&chunk, sizeof(chunk), 1)) == 1)
					{
						result = ProcessCOBChunk(fptr, chunk, cobContext, pObject, pDebug);
						if (result == COBCHUNK_ERROR)
						{
							// set flag to delete pObject
							read_result = 0;							
							break;
						} else if (result == COBCHUNK_EOF)
							break;
					}
					if (read_result != 1)
					{
						pObject = NULL;

						if (pDebug != NULL)
							pDebug->ErrorOut("COB file error: EOF reached before END chunk %s\n", (const char*)lpszFileName);
					}
				} else
				{
					if (pDebug != NULL)
						pDebug->ErrorOut("Not a binary COB file %s\n", (const char*)lpszFileName);
				}
			} else
			{
				if (pDebug != NULL)
					pDebug->ErrorOut("Not a version 00.01 Caligari COB file %s\n", (const char*)lpszFileName);
			}
		} else
		{
			if (pDebug != NULL)
				pDebug->ErrorOut("Abnormal COB file header termination %s\n", (const char*)lpszFileName);
		}

		fptr->CloseFile();

		return (pObject == (Pointer<SBObject>)this);
	} else // if (fptr != NULL)
	{
		if (pDebug != NULL)
			pDebug->ErrorOut("Unable to open COB file %s\n", (const char*)lpszFileName);
		return false;
	}
}

