/***************************************
* File    : SBObjectClass.h
* Author  : Shezan Baig 
* Date    : 5/3/2001
* Purpose : Declares a concrete class for SB3DObject interface
***************************************/

#ifndef _SBOBJECTCLASS_H_INCLUDED
#define _SBOBJECTCLASS_H_INCLUDED


// typedefs for 3D data types
typedef float SBcoord[2];
typedef float SBvertex[3];
typedef float SBvector[3];
typedef float SBrgba[4];

/**********************************************
* Function      : VectorsEqual
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Check whether two vectors are pointing in the
*				  same direction
* Return Value  : (bool) true if yes, false if no
* Parameters    : The two SBvector to be compared
**********************************************/
inline bool VectorsEqual(const SBvector t1, const SBvector t2)
{
	return (t1[SB_X] == t2[SB_X]
			&& t1[SB_Y] == t2[SB_Y]
			&& t1[SB_Z] == t2[SB_Z]);
}

/**********************************************
* Function      : CopyVector
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Copies the values from one vector to another
* Return Value  : None
* Parameters    : (SBvector) t1 - the destination
*				  (SBvector) t2 - the source
**********************************************/
inline void CopyVector(SBvector t1, const SBvector t2)
{
	t1[SB_X] = t2[SB_X];
	t1[SB_Y] = t2[SB_Y];
	t1[SB_Z] = t2[SB_Z];
}

/**********************************************
* Function      : VerticesEqual
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Check whether two vertices are at the same location
* Return Value  : (SBbool) true if yes, false if no
* Parameters    : The two SBvertex to be compared
**********************************************/
inline bool VerticesEqual(const SBvertex t1, const SBvertex t2)
{
	return (t1[SB_X] == t2[SB_X]
			&& t1[SB_Y] == t2[SB_Y]
			&& t1[SB_Z] == t2[SB_Z]);
}

/**********************************************
* Function      : CopyVertex
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Copies the values from one vertex to another
* Return Value  : None
* Parameters    : (SBvertex) t1 - the destination
*				  (SBvertex) t2 - the source
**********************************************/
inline void CopyVertex(SBvertex t1, const SBvertex t2)
{
	t1[SB_X] = t2[SB_X];
	t1[SB_Y] = t2[SB_Y];
	t1[SB_Z] = t2[SB_Z];
}

/**********************************************
* Function      : TranslateVertex
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Moves the given vertex in x,y,z direction by the
*				  given amount
* Return Value  : None
* Parameters    : (SBvertex) t1 - the vertex to be moved
*				  (SBfloat) xAmt - amount to move in x direction
*				  (SBfloat) yAmt - amount to move in y direction
*				  (SBfloat) zAmt - amount to move in z direction
**********************************************/
inline void TranslateVertex(SBvertex t1, const float& xAmt, const float& yAmt, const float& zAmt)
{
	t1[SB_X] = t1[SB_X] + xAmt;
	t1[SB_Y] = t1[SB_Y] + yAmt;
	t1[SB_Z] = t1[SB_Z] + zAmt;
}

/**********************************************
* Function      : TranslateVertexV
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : Moves the given vertex in x,y,z direction by the
*				  given amount in the vector
* Return Value  : None
* Parameters    : (SBvertex) t1 - the vertex to be moved
*				  (SBvector) v1 - amount to move in x,y,z direction
**********************************************/
inline void TranslateVertexV(SBvertex t1, const SBvector v1)
{
	t1[SB_X] = t1[SB_X] + v1[SB_X];
	t1[SB_Y] = t1[SB_Y] + v1[SB_Y];
	t1[SB_Z] = t1[SB_Z] + v1[SB_Z];
}

#include <math.h>
/**********************************************
* Function      : ReduceToUnit
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : makes the given vertex a unit vertex, this
*				  means its length is one unit
* Return Value  : None
* Parameters    : (SBvector) v - the vector to be adjusted
**********************************************/
inline void ReduceToUnit(SBvector v)
{
	float length = (float)sqrt(v[0]*v[0]
							+ v[1]*v[1]
							+ v[2]*v[2]);

	if (length == 0.0f)
		return;

	float length_inverse = 1.0f/length;

	v[0] *= length_inverse;
	v[1] *= length_inverse;
	v[2] *= length_inverse;
}

/**********************************************
* Function      : CalcNormal
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : calculates the normal vector for three vertices
*				  the vertices should be provided in counter-clockwise
*				  direction for the out vector to point upwards from
*				  the surface
* Return Value  : None
* Parameters    : (SBvertex) vertice1 - first vertex
*				  (SBvertex) vertice2 - second vertex
*				  (SBvertex) vertice3 - third vertex
*				  (SBvector) out - resulting normal vector
**********************************************/
inline void CalcNormal(const SBvertex vertice1, const SBvertex vertice2, const SBvertex vertice3, SBvector out)
{
	SBvertex v1, v2;

	v1[0] = vertice1[0] - vertice2[0];
	v1[1] = vertice1[1] - vertice2[1];
	v1[2] = vertice1[2] - vertice2[2];

	v2[0] = vertice2[0] - vertice3[0];
	v2[1] = vertice2[1] - vertice3[1];
	v2[2] = vertice2[2] - vertice3[2];


	out[0] = v1[1]*v2[2] - v1[2]*v2[1];
	out[1] = v1[2]*v2[0] - v1[0]*v2[2];
	out[2] = v1[0]*v2[1] - v1[1]*v2[0];


	ReduceToUnit(out);
}

/**********************************************
* Function      : DotProduct
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : find out the dot product of two vectors (should be unit length)
* Return Value  : (SBfloat) the dot product
* Parameters    : (SBvector) v1 - the first vector
*				  (SBvector) v2 - the second vector
**********************************************/
inline float DotProduct(const SBvector v1, const SBvector v2)
{
	return ((v1[0]*v2[0])
			+ (v1[1]*v2[1])
			+ (v1[2]*v2[2]));
}


/**********************************************
* Function      : AngleBetween
* Author        : Shezan Baig 
* Date          : 5/3/2001
* Purpose       : get the angle between two vectors (should be unit length)
* Return Value  : (float) the angle (in degrees)
* Parameters    : (SBvector) v1 - the first vector
*				  (SBvector) v2 - the second vector
**********************************************/
inline float AngleBetween(const SBvector v1, const SBvector v2)
{
	return (float)RTOD( ((float)acos(DotProduct(v1,v2))) );
}




#include <vector>

// defines a material
typedef struct
{
	// the color
	SBrgba m_rgbaColor;
	// texture index (usage to be determined later)
	int m_nTexture;
	// ambience
	float m_fAmbience;
	// shininess
	float m_fSpecular;
	// roughness
//	float m_fRoughness;
} Tag_Material;

// defines a three sided face (triangle)
typedef struct
{
	// index to vertices in SBObject::m_pVertices
	int m_nVerts[3];
	// index to texture coordinates in SBObject::m_pTextures
	int m_nTextureCoords[3];
	// index to normal vector in SBObject::m_pNormals
	// one normal for each vertex
	int m_nNormals[3];
	// index to material in SBObject::m_pMaterials
	int m_nMaterial;
} Tag_Face;


// this temporary structure is required to store SBvector objects in
// an std::vector variable
typedef struct
{
	SBvector v;
} Tag_TmpVector;

#include "LoadCOB.h"

class SBObject : public virtual BaseObject,
					public SB3DObject
{
	DECLARE_RTTI(SBObject)
	BEGIN_INTERFACE_LIST(SBObject)
		INTERFACE(SB3DObject)
		INTERFACE(SBObject)
	END_INTERFACE_LIST
public:
	// default constructor
	SBObject();
	// destructor
	virtual ~SBObject();

// Methods inherited from SB3DObject interface
	// load 3D object from a file, should return true if successful
	// the SBFilePtr interface can be supplied by the caller, or set
	// to NULL to use the default DiskFile
	virtual bool LoadFromFile(SBFilePtr r_pFile, const String& lpszFileName);
	// save 3D object to a file, should return true if successful
	// the SBFilePtr interface can be supplied by the caller, or set
	// to NULL to use the default DiskFile
	virtual bool SaveToFile(SBFilePtr r_pFile, const String& lpszFileName) const;
	// call GL functions for drawing object
	virtual void Display() const;
	// puts this 3D object in a display list
	// returns the name of the display list
	virtual unsigned int CreateDisplayList();

	// gets the sub-object with the specified name
	virtual SB3DObjectPtr GetObjectFromName(const String& lpszName) const;
	// get the name of the this object
	virtual String GetName() const;
	// sets the position of the object
	virtual void SetPosition(Float3CTPtr vPosition);
	// sets the direction to be pointed to
	virtual void SetDirection(Float3CTPtr vDirection);
	// gets the position of the object
	virtual Float3CTPtr GetPosition() const;
	// gets the direction of the object
	virtual Float3CTPtr GetDirection() const;

	// sets the rotation of the object
	virtual void SetRotation(Float3CTPtr afRotation);
	// gets the rotation of the object
	virtual Float3CTPtr GetRotation() const;
	// sets the scale of the object
	virtual void SetScale(Float3CTPtr afScale);
	// gets the scale of the object
	virtual Float3CTPtr GetScale() const;

	// adjusts vertex normals to smooth out edges with angles <=
	// the tolerance angle provided
	virtual void SmoothSurfaces(float fToleranceAngle);



private:
	// checks whether the two faces share edges or not
	inline bool ShareEdges(int nFace1, int nFace2);

	inline void SetVerticeCount(int nCount)
	{
		if (m_pVertices != NULL)
		{
			delete[] m_pVertices;
			m_pVertices = NULL;
		}
		if (nCount > 0)
			m_pVertices = new SBvertex[nCount];
		m_nVertices = nCount;
	}
	inline void SetTextureCoordCount(int nCount)
	{
		if (m_pTextures != NULL)
		{
			delete[] m_pTextures;
			m_pTextures = NULL;
		}
		if (nCount > 0)
			m_pTextures = new SBcoord[nCount];
		m_nTextures = nCount;
	}
	inline void SetNormalCount(int nCount)
	{
		if (m_pNormals != NULL)
		{
			delete[] m_pNormals;
			m_pNormals = NULL;
		}
		if (nCount > 0)
			m_pNormals = new SBvector[nCount];
		m_nNormals = nCount;
	}
	inline void SetMaterialCount(int nCount)
	{
		if (m_pMaterials != NULL)
		{
			delete[] m_pMaterials;
			m_pMaterials = NULL;
		}
		if (nCount > 0)
			m_pMaterials = new Tag_Material[nCount];
		m_nMaterials = nCount;
	}
	inline void SetMaterials(int nCount, Tag_Material* pData)
	{
		if (m_pMaterials != NULL)
		{
			delete[] m_pMaterials;
		}
		m_pMaterials = pData;
		m_nMaterials = nCount;
	}
	inline void SetFaceCount(int nCount)
	{
		if (m_pFaces != NULL)
		{
			delete[] m_pFaces;
			m_pFaces = NULL;
		}
		if (nCount > 0)
			m_pFaces = new Tag_Face[nCount];
		m_nFaces = nCount;
	}

	inline int GetVerticeCount()
	{
		return m_nVertices;
	}
	inline int GetTextureCoordCount()
	{
		return m_nTextures;
	}
	inline int GetNormalCount()
	{
		return m_nNormals;
	}
	inline int GetMaterialCount()
	{
		return m_nMaterials;
	}
	inline int GetFaceCount()
	{
		return m_nFaces;
	}

	Array< Pointer<SBObject> > m_aSubObjects;
	SBObject* m_pParentObject;

	// name of display list (0 if display list not created)
	GLuint m_nDisplayList;

	// position of object (based on axis)
	SBvertex m_vPosition;

	// direction pointed to by the x,y and z axes
	SBvector m_avAxes[3];

	// rotation on SB_X,SB_Y,SB_Z axes
	float m_afRotation[3];

	// scale on SB_X,SB_Y,SB_Z axes
	float m_afScale[3];

	// texture coordinates
	SBcoord* m_pTextures;
	// number of texture coordinates
	int m_nTextures;

	// array of vertices for this object
	SBvertex* m_pVertices;
	// number of vertices
	int m_nVertices;

	// array of normal vectors for this object
	SBvector* m_pNormals;
	// number of normals
	int m_nNormals;

	// array of materials used in this object
	Tag_Material* m_pMaterials;
	// number of materials
	int m_nMaterials;

	// array of faces for this object
	Tag_Face* m_pFaces;
	// number of faces
	int m_nFaces;

	// name of the object
	char* m_sName;

	// set to true while display list is being created
	bool m_bCreatingDL;


// LoadCOB functions
	static int ProcessCOBPolHChunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug);
	static int ProcessCOBGrouChunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug);
	static int ProcessCOBMat1Chunk(SBFilePtr fptr, Pointer<SBObject> pParentObject, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, DebuggerCTPtr pDebug);
	static int ProcessCOBChunk(SBFilePtr fptr, Tag_COBChunkHeader& hdr, Tag_COBContext& cobContext, Pointer<SBObject> pObject, DebuggerCTPtr pDebug);
	bool LoadCOBFile(SBFilePtr fptr, const String& lpszFileName, DebuggerCTPtr pDebug);
};
typedef Pointer<SBObject> SBObjectPtr;
typedef CTPointer<SBObject> SBObjectCTPtr;



#endif // _SBOBJECTCLASS_H_INCLUDED
