/***************************************
* File    : Win32Impl.h
* Author  : Shezan Baig 
* Date    : 2/3/2001
* Purpose : Header file for CWin32Impl class
***************************************/

#ifndef _WIN32IMPL_H_INCLUDED
#define _WIN32IMPL_H_INCLUDED

#define LEFT_BUTTON		0
#define RIGHT_BUTTON	1
#define MIDDLE_BUTTON	2

/**********************************************
* Class         : CWin32Impl
* Author        : Shezan Baig 
* Date          : 2/3/2001
* Purpose       : Implementation of platform interfaces for Win32
**********************************************/
class CWin32Impl : public virtual BaseObject,
					public Window,
					public Input,
					public Timer
{
	/* Customised DECLARE_RTTI */
private:
	static RTTIClass rtti_for_CWin32Impl;
	static void* rtti_CreateCWin32Impl(const char* r_sInterface)
	{
		if (CWin32Impl::OneAndOnlyWin32Impl == NULL)
			CWin32Impl::OneAndOnlyWin32Impl = new CWin32Impl;

		return CWin32Impl::OneAndOnlyWin32Impl->FindInterface(r_sInterface);
	}

	/* implementation of RTTI interface */
public:
	virtual const RTTIClass* GetClass() const
	{
		return &rtti_for_CWin32Impl;
	}
	static bool rtti_ImplementsInterfaceCWin32Impl(const char* r_sInterface)
	{										
		return false;
	}											

	BEGIN_INTERFACE_LIST(CWin32Impl)
		INTERFACE(Window)
		INTERFACE(Input)
		INTERFACE(Timer)
		INTERFACE(CWin32Impl)
	END_INTERFACE_LIST

	friend LONG WINAPI WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
public:
	CWin32Impl()
	{
		m_hWnd = NULL;
		m_hPalette = NULL;
		m_bSettingMousePos = false;
		m_bFullScreenMode = false;

		m_hRCSema = CreateSemaphore(NULL, 1, 1, NULL);

		GFC::GetDebugger()->DebugOut("CWin32Impl constructed\n");
	}
	virtual ~CWin32Impl()
	{
		CloseHandle(m_hRCSema);
		m_hRCSema = NULL;

		GFC::GetDebugger()->DebugOut("* CWin32Impl destructed\n");
	}

// functions from Window
	virtual bool CreateWindow(const String& lpszTitle);
	virtual bool CloseWindow();
	virtual bool ShowWindow();

	virtual bool SetFullScreen();
	virtual bool SetWindowRect(int left, int top, int right, int bottom);

	virtual bool FlushMessages();

	// sets the pointer to the IOpenGL interface for callback functions
	virtual void SetOpenGL(OGLPtr lpOpenGL);

	// invalidates the window and forces repaint
	virtual void InvalidateWindow();

	// this function will only work under Win32
	virtual HWND GetHwnd()
	{
		return m_hWnd;
	}
	virtual HDC GetHdc()
	{
		return m_hDC;
	}
	virtual void CaptureRC()
	{
		WaitForSingleObject(m_hRCSema, INFINITE);
		wglMakeCurrent(m_hDC, m_hGLRC);
	}
	virtual void ReleaseRC()
	{
		wglMakeCurrent(NULL, NULL);
		ReleaseSemaphore(m_hRCSema, 1, NULL);
	}


	// this function will only work if OpenGL is setup
	virtual HGLRC GetHglRC()
	{
		return m_hGLRC;
	}


// functions for Input
	virtual bool IsKeyDown(int nKeyCode) const;
	virtual bool IsLButtonDown() const ;
	virtual bool IsRButtonDown() const;

	virtual void SetInputTarget(InputTargetPtr lpTarget);

	// sets the position of the mouse to the ones specified by nX and xY
	// if nX and nY are -1, then the mouse is centered in the window, and
	// nX and nY are updated accordingly
	virtual void SetMousePosition(int& nX, int& nY);

// functions for Timer
	virtual unsigned long GetTickCount() const
	{
		return ::GetTickCount();
	}

private:
// helper functions
	void SetDCPixelFormat(HDC hdc);
	void InitialiseInputData()
	{
		m_bMouseState[LEFT_BUTTON] = false;
		m_bMouseState[RIGHT_BUTTON] = false;
		m_bMouseState[MIDDLE_BUTTON] = false;

		int cnt;
		for (cnt = 0; cnt < 257; cnt++)
			m_bKeyState[cnt] = false;
	}

	HWND m_hWnd;
	HGLRC m_hGLRC;
	HDC m_hDC;
	HANDLE m_hRCSema;

	InputTargetPtr m_pInputTarget;
	OGLPtr m_pOpenGL;

	bool m_bSettingMousePos; // set to true when SetMousePosition is called
	bool m_bMouseState[3];
	bool m_bKeyState[257];

	bool m_bFullScreenMode;

	HPALETTE m_hPalette;

	static Pointer<CWin32Impl> OneAndOnlyWin32Impl;
};

#endif // #ifndef _WIN32IMPL_H_INCLUDED
